from __future__ import print_function, absolute_import, unicode_literals
import wormhole
from twisted.internet import reactor
from twisted.internet.defer import Deferred, inlineCallbacks, gatherResults
from twisted.internet.protocol import Protocol, Factory
from twisted.trial import unittest

from ..common import ServerBase, poll_until
from ..._interfaces import IDilationConnector
from ...eventual import EventualQueue
from ..._dilation._noise import NoiseConnection

APPID = u"lothar.com/dilate-test"


def doBoth(d1, d2):
    return gatherResults([d1, d2], True)


class L(Protocol):
    def connectionMade(self):
        print("got connection")
        self.transport.write(b"hello\n")

    def dataReceived(self, data):
        print("dataReceived: {}".format(data))
        self.factory.d.callback(data)

    def connectionLost(self, why):
        print("connectionLost")


class Full(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        if not NoiseConnection:
            raise unittest.SkipTest("noiseprotocol unavailable")
        # test_welcome wants to see [current_cli_version]
        yield self._setup_relay(None)

    @inlineCallbacks
    def test_control(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w1.allocate_code()
        code = yield w1.get_code()
        print("code is: {}".format(code))
        w2.set_code(code)
        yield doBoth(w1.get_verifier(), w2.get_verifier())
        print("connected")

        eps1 = w1.dilate()
        eps2 = w2.dilate()
        print("w.dilate ready")

        f1 = Factory()
        f1.protocol = L
        f1.d = Deferred()
        f1.d.addCallback(lambda data: eq.fire_eventually(data))
        d1 = eps1.control.connect(f1)

        f2 = Factory()
        f2.protocol = L
        f2.d = Deferred()
        f2.d.addCallback(lambda data: eq.fire_eventually(data))
        d2 = eps2.control.connect(f2)
        yield d1
        yield d2
        print("control endpoints connected")
        # note: I'm making some horrible assumptions about one-to-one writes
        # and reads across a TCP stack that isn't obligated to maintain such
        # a relationship, but it's much easier than doing this properly. If
        # the tests ever start failing, do the extra work, probably by
        # using a twisted.protocols.basic.LineOnlyReceiver
        data1 = yield f1.d
        data2 = yield f2.d
        self.assertEqual(data1, b"hello\n")
        self.assertEqual(data2, b"hello\n")

        yield w1.close()
        yield w2.close()
    test_control.timeout = 30


class ReconP(Protocol):
    def eventually(self, which, data):
        d = self.factory.deferreds[which]
        self.factory.eq.fire_eventually(data).addCallback(d.callback)

    def connectionMade(self):
        self.eventually("connectionMade", self)
        # self.transport.write(b"hello\n")

    def dataReceived(self, data):
        self.eventually("dataReceived", data)

    def connectionLost(self, why):
        self.eventually("connectionLost", (self, why))


class ReconF(Factory):
    protocol = ReconP

    def __init__(self, eq):
        Factory.__init__(self)
        self.eq = eq
        self.deferreds = {}
        for name in ["connectionMade", "dataReceived", "connectionLost"]:
            self.deferreds[name] = Deferred()

    def resetDeferred(self, name):
        d = Deferred()
        self.deferreds[name] = d
        return d


class Reconnect(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        if not NoiseConnection:
            raise unittest.SkipTest("noiseprotocol unavailable")
        # test_welcome wants to see [current_cli_version]
        yield self._setup_relay(None)

    @inlineCallbacks
    def test_reconnect(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        yield doBoth(w1.get_verifier(), w2.get_verifier())

        eps1 = w1.dilate()
        eps2 = w2.dilate()
        print("w.dilate ready")

        f1, f2 = ReconF(eq), ReconF(eq)
        d1, d2 = eps1.control.connect(f1), eps2.control.connect(f2)
        yield d1
        yield d2

        protocols = {}

        def p_connected(p, index):
            protocols[index] = p
            msg = "hello from %s\n" % index
            p.transport.write(msg.encode("ascii"))
        f1.deferreds["connectionMade"].addCallback(p_connected, 1)
        f2.deferreds["connectionMade"].addCallback(p_connected, 2)

        data1 = yield f1.deferreds["dataReceived"]
        data2 = yield f2.deferreds["dataReceived"]
        self.assertEqual(data1, b"hello from 2\n")
        self.assertEqual(data2, b"hello from 1\n")
        # the ACKs are now in flight and may not arrive before we kill the
        # connection

        f1.resetDeferred("connectionMade")
        f2.resetDeferred("connectionMade")
        d1 = f1.resetDeferred("dataReceived")
        d2 = f2.resetDeferred("dataReceived")

        # now we reach inside and drop the connection
        sc = protocols[1].transport
        orig_connection = sc._manager._connection
        orig_connection.disconnect()

        # stall until the connection has been replaced
        yield poll_until(lambda: sc._manager._connection
                         and (orig_connection != sc._manager._connection))

        # now write some more data, which should travel over the new
        # connection
        protocols[1].transport.write(b"more\n")
        data2 = yield d2
        self.assertEqual(data2, b"more\n")

        replacement_connection = sc._manager._connection
        self.assertNotEqual(orig_connection, replacement_connection)

        # the application-visible Protocol should not observe the
        # interruption
        self.assertNoResult(f1.deferreds["connectionMade"])
        self.assertNoResult(f2.deferreds["connectionMade"])
        self.assertNoResult(f1.deferreds["connectionLost"])
        self.assertNoResult(f2.deferreds["connectionLost"])

        yield w1.close()
        yield w2.close()

    @inlineCallbacks
    def test_data_while_offline(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        yield doBoth(w1.get_verifier(), w2.get_verifier())

        eps1 = w1.dilate()
        eps2 = w2.dilate()
        print("w.dilate ready")

        f1, f2 = ReconF(eq), ReconF(eq)
        d1, d2 = eps1.control.connect(f1), eps2.control.connect(f2)
        yield d1
        yield d2

        protocols = {}

        def p_connected(p, index):
            protocols[index] = p
            msg = "hello from %s\n" % index
            p.transport.write(msg.encode("ascii"))
        f1.deferreds["connectionMade"].addCallback(p_connected, 1)
        f2.deferreds["connectionMade"].addCallback(p_connected, 2)

        data1 = yield f1.deferreds["dataReceived"]
        data2 = yield f2.deferreds["dataReceived"]
        self.assertEqual(data1, b"hello from 2\n")
        self.assertEqual(data2, b"hello from 1\n")
        # the ACKs are now in flight and may not arrive before we kill the
        # connection

        f1.resetDeferred("connectionMade")
        f2.resetDeferred("connectionMade")
        d1 = f1.resetDeferred("dataReceived")
        d2 = f2.resetDeferred("dataReceived")

        # switch off connections
        assert not w1._boss._D._manager._debug_stall_connector
        cd1, cd2 = Deferred(), Deferred()
        w1._boss._D._manager._debug_stall_connector = cd1.callback
        w2._boss._D._manager._debug_stall_connector = cd2.callback

        # now we reach inside and drop the connection
        sc = protocols[1].transport
        orig_connection = sc._manager._connection
        orig_connection.disconnect()

        c1 = yield cd1
        c2 = yield cd2
        assert IDilationConnector.providedBy(c1)
        assert IDilationConnector.providedBy(c2)
        assert c1 is not orig_connection
        w1._boss._D._manager._debug_stall_connector = False
        w2._boss._D._manager._debug_stall_connector = False

        # now write some data while the connection is definitely offline
        protocols[1].transport.write(b"more 1->2\n")
        protocols[2].transport.write(b"more 2->1\n")

        # allow the connections to proceed
        c1.start()
        c2.start()

        # and wait for the data to arrive
        data2 = yield d2
        self.assertEqual(data2, b"more 1->2\n")
        data1 = yield d1
        self.assertEqual(data1, b"more 2->1\n")

        # the application-visible Protocol should not observe the
        # interruption
        self.assertNoResult(f1.deferreds["connectionMade"])
        self.assertNoResult(f2.deferreds["connectionMade"])
        self.assertNoResult(f1.deferreds["connectionLost"])
        self.assertNoResult(f2.deferreds["connectionLost"])

        yield w1.close()
        yield w2.close()


class Endpoints(ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        if not NoiseConnection:
            raise unittest.SkipTest("noiseprotocol unavailable")
        # test_welcome wants to see [current_cli_version]
        yield self._setup_relay(None)

    @inlineCallbacks
    def test_endpoints(self):
        eq = EventualQueue(reactor)
        w1 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w2 = wormhole.create(APPID, self.relayurl, reactor, _enable_dilate=True)
        w1.allocate_code()
        code = yield w1.get_code()
        w2.set_code(code)
        yield doBoth(w1.get_verifier(), w2.get_verifier())

        eps1 = w1.dilate()
        eps2 = w2.dilate()
        print("w.dilate ready")

        f0 = ReconF(eq)
        yield eps2.listen.listen(f0)

        from twisted.python import log
        f1 = ReconF(eq)
        log.msg("connecting")
        p1_client = yield eps1.connect.connect(f1)
        log.msg("sending c->s")
        p1_client.transport.write(b"hello from p1\n")
        data = yield f0.deferreds["dataReceived"]
        self.assertEqual(data, b"hello from p1\n")
        p1_server = self.successResultOf(f0.deferreds["connectionMade"])
        log.msg("sending s->c")
        p1_server.transport.write(b"hello p1\n")
        log.msg("waiting for client to receive")
        data = yield f1.deferreds["dataReceived"]
        self.assertEqual(data, b"hello p1\n")

        # open a second channel
        f0.resetDeferred("connectionMade")
        f0.resetDeferred("dataReceived")
        f1.resetDeferred("dataReceived")
        f2 = ReconF(eq)
        p2_client = yield eps1.connect.connect(f2)
        p2_server = yield f0.deferreds["connectionMade"]
        p2_server.transport.write(b"hello p2\n")
        data = yield f2.deferreds["dataReceived"]
        self.assertEqual(data, b"hello p2\n")
        p2_client.transport.write(b"hello from p2\n")
        data = yield f0.deferreds["dataReceived"]
        self.assertEqual(data, b"hello from p2\n")
        self.assertNoResult(f1.deferreds["dataReceived"])

        # now close the first subchannel (p1) from the listener side
        p1_server.transport.loseConnection()
        yield f0.deferreds["connectionLost"]
        yield f1.deferreds["connectionLost"]

        f0.resetDeferred("connectionLost")
        # and close the second from the connector side
        p2_client.transport.loseConnection()
        yield f0.deferreds["connectionLost"]
        yield f2.deferreds["connectionLost"]

        yield w1.close()
        yield w2.close()



from __future__ import absolute_import, print_function, unicode_literals

from itertools import count

from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.threads import deferToThread
from twisted.trial import unittest

from unittest import mock

from .._rlcompleter import (CodeInputter, _input_code_with_completion,
                            input_with_completion, warn_readline)
from ..errors import AlreadyInputNameplateError, KeyFormatError

APPID = "appid"


class Input(unittest.TestCase):
    @inlineCallbacks
    def test_wrapper(self):
        helper = object()
        trueish = object()
        with mock.patch(
                "wormhole._rlcompleter._input_code_with_completion",
                return_value=trueish) as m:
            used_completion = yield input_with_completion(
                "prompt:", helper, reactor)
        self.assertIs(used_completion, trueish)
        self.assertEqual(m.mock_calls, [mock.call("prompt:", helper, reactor)])
        # note: if this test fails, the warn_readline() message will probably
        # get written to stderr


class Sync(unittest.TestCase):
    # exercise _input_code_with_completion, which uses the blocking builtin
    # "input()" function, hence _input_code_with_completion is usually in a
    # thread with deferToThread

    @mock.patch("wormhole._rlcompleter.CodeInputter")
    @mock.patch("wormhole._rlcompleter.readline", __doc__="I am GNU readline")
    @mock.patch("wormhole._rlcompleter.input", return_value="code")
    def test_readline(self, input, readline, ci):
        c = mock.Mock(name="inhibit parenting")
        c.completer = object()
        trueish = object()
        c.used_completion = trueish
        ci.configure_mock(return_value=c)
        prompt = object()
        input_helper = object()
        reactor = object()
        used = _input_code_with_completion(prompt, input_helper, reactor)
        self.assertIs(used, trueish)
        self.assertEqual(ci.mock_calls, [mock.call(input_helper, reactor)])
        self.assertEqual(c.mock_calls, [mock.call.finish("code")])
        self.assertEqual(input.mock_calls, [mock.call(prompt)])
        self.assertEqual(readline.mock_calls, [
            mock.call.parse_and_bind("tab: complete"),
            mock.call.set_completer(c.completer),
            mock.call.set_completer_delims(""),
        ])

    @mock.patch("wormhole._rlcompleter.CodeInputter")
    @mock.patch("wormhole._rlcompleter.readline")
    @mock.patch("wormhole._rlcompleter.input", return_value="code")
    def test_readline_no_docstring(self, input, readline, ci):
        del readline.__doc__  # when in doubt, it assumes GNU readline
        c = mock.Mock(name="inhibit parenting")
        c.completer = object()
        trueish = object()
        c.used_completion = trueish
        ci.configure_mock(return_value=c)
        prompt = object()
        input_helper = object()
        reactor = object()
        used = _input_code_with_completion(prompt, input_helper, reactor)
        self.assertIs(used, trueish)
        self.assertEqual(ci.mock_calls, [mock.call(input_helper, reactor)])
        self.assertEqual(c.mock_calls, [mock.call.finish("code")])
        self.assertEqual(input.mock_calls, [mock.call(prompt)])
        self.assertEqual(readline.mock_calls, [
            mock.call.parse_and_bind("tab: complete"),
            mock.call.set_completer(c.completer),
            mock.call.set_completer_delims(""),
        ])

    @mock.patch("wormhole._rlcompleter.CodeInputter")
    @mock.patch("wormhole._rlcompleter.readline", __doc__="I am libedit")
    @mock.patch("wormhole._rlcompleter.input", return_value="code")
    def test_libedit(self, input, readline, ci):
        c = mock.Mock(name="inhibit parenting")
        c.completer = object()
        trueish = object()
        c.used_completion = trueish
        ci.configure_mock(return_value=c)
        prompt = object()
        input_helper = object()
        reactor = object()
        used = _input_code_with_completion(prompt, input_helper, reactor)
        self.assertIs(used, trueish)
        self.assertEqual(ci.mock_calls, [mock.call(input_helper, reactor)])
        self.assertEqual(c.mock_calls, [mock.call.finish("code")])
        self.assertEqual(input.mock_calls, [mock.call(prompt)])
        self.assertEqual(readline.mock_calls, [
            mock.call.parse_and_bind("bind ^I rl_complete"),
            mock.call.set_completer(c.completer),
            mock.call.set_completer_delims(""),
        ])

    @mock.patch("wormhole._rlcompleter.CodeInputter")
    @mock.patch("wormhole._rlcompleter.readline", None)
    @mock.patch("wormhole._rlcompleter.input", return_value="code")
    def test_no_readline(self, input, ci):
        c = mock.Mock(name="inhibit parenting")
        c.completer = object()
        trueish = object()
        c.used_completion = trueish
        ci.configure_mock(return_value=c)
        prompt = object()
        input_helper = object()
        reactor = object()
        used = _input_code_with_completion(prompt, input_helper, reactor)
        self.assertIs(used, trueish)
        self.assertEqual(ci.mock_calls, [mock.call(input_helper, reactor)])
        self.assertEqual(c.mock_calls, [mock.call.finish("code")])
        self.assertEqual(input.mock_calls, [mock.call(prompt)])

    @mock.patch("wormhole._rlcompleter.CodeInputter")
    @mock.patch("wormhole._rlcompleter.readline", None)
    @mock.patch("wormhole._rlcompleter.input", return_value=b"code")
    def test_bytes(self, input, ci):
        c = mock.Mock(name="inhibit parenting")
        c.completer = object()
        trueish = object()
        c.used_completion = trueish
        ci.configure_mock(return_value=c)
        prompt = object()
        input_helper = object()
        reactor = object()
        used = _input_code_with_completion(prompt, input_helper, reactor)
        self.assertIs(used, trueish)
        self.assertEqual(ci.mock_calls, [mock.call(input_helper, reactor)])
        self.assertEqual(c.mock_calls, [mock.call.finish(u"code")])
        self.assertEqual(input.mock_calls, [mock.call(prompt)])


def get_completions(c, prefix):
    completions = []
    for state in count(0):
        text = c.completer(prefix, state)
        if text is None:
            return completions
        completions.append(text)


def fake_blockingCallFromThread(f, *a, **kw):
    return f(*a, **kw)


class Completion(unittest.TestCase):
    def test_simple(self):
        # no actual completion
        helper = mock.Mock()
        c = CodeInputter(helper, "reactor")
        c.bcft = fake_blockingCallFromThread
        c.finish("1-code-ghost")
        self.assertFalse(c.used_completion)
        self.assertEqual(helper.mock_calls, [
            mock.call.choose_nameplate("1"),
            mock.call.choose_words("code-ghost")
        ])

    @mock.patch(
        "wormhole._rlcompleter.readline",
        get_completion_type=mock.Mock(return_value=0))
    def test_call(self, readline):
        # check that it calls _commit_and_build_completions correctly
        helper = mock.Mock()
        c = CodeInputter(helper, "reactor")
        c.bcft = fake_blockingCallFromThread

        # pretend nameplates: 1, 12, 34

        # first call will be with "1"
        cabc = mock.Mock(return_value=["1", "12"])
        c._commit_and_build_completions = cabc

        self.assertEqual(get_completions(c, "1"), ["1", "12"])
        self.assertEqual(cabc.mock_calls, [mock.call("1")])

        # then "12"
        cabc.reset_mock()
        cabc.configure_mock(return_value=["12"])
        self.assertEqual(get_completions(c, "12"), ["12"])
        self.assertEqual(cabc.mock_calls, [mock.call("12")])

        # now we have three "a" words: "and", "ark", "aaah!zombies!!"
        cabc.reset_mock()
        cabc.configure_mock(return_value=["aargh", "ark", "aaah!zombies!!"])
        self.assertEqual(
            get_completions(c, "12-a"), ["aargh", "ark", "aaah!zombies!!"])
        self.assertEqual(cabc.mock_calls, [mock.call("12-a")])

        cabc.reset_mock()
        cabc.configure_mock(return_value=["aargh", "aaah!zombies!!"])
        self.assertEqual(
            get_completions(c, "12-aa"), ["aargh", "aaah!zombies!!"])
        self.assertEqual(cabc.mock_calls, [mock.call("12-aa")])

        cabc.reset_mock()
        cabc.configure_mock(return_value=["aaah!zombies!!"])
        self.assertEqual(get_completions(c, "12-aaa"), ["aaah!zombies!!"])
        self.assertEqual(cabc.mock_calls, [mock.call("12-aaa")])

        c.finish("1-code")
        self.assert_(c.used_completion)

    def test_wrap_error(self):
        helper = mock.Mock()
        c = CodeInputter(helper, "reactor")
        c._wrapped_completer = mock.Mock(side_effect=ValueError("oops"))
        with mock.patch("wormhole._rlcompleter.traceback") as traceback:
            with mock.patch("wormhole._rlcompleter.print") as mock_print:
                with self.assertRaises(ValueError) as e:
                    c.completer("text", 0)
        self.assertEqual(traceback.mock_calls, [mock.call.print_exc()])
        self.assertEqual(mock_print.mock_calls,
                         [mock.call("completer exception: oops")])
        self.assertEqual(str(e.exception), "oops")

    @inlineCallbacks
    def test_build_completions(self):
        rn = mock.Mock()
        # InputHelper.get_nameplate_completions returns just the suffixes
        gnc = mock.Mock()  # get_nameplate_completions
        cn = mock.Mock()  # choose_nameplate
        gwc = mock.Mock()  # get_word_completions
        cw = mock.Mock()  # choose_words
        helper = mock.Mock(
            refresh_nameplates=rn,
            get_nameplate_completions=gnc,
            choose_nameplate=cn,
            get_word_completions=gwc,
            choose_words=cw,
        )
        # this needs a real reactor, for blockingCallFromThread
        c = CodeInputter(helper, reactor)
        cabc = c._commit_and_build_completions

        # in this test, we pretend that nameplates 1,12,34 are active.

        # 43 TAB -> nothing (and refresh_nameplates)
        gnc.configure_mock(return_value=[])
        matches = yield deferToThread(cabc, "43")
        self.assertEqual(matches, [])
        self.assertEqual(rn.mock_calls, [mock.call()])
        self.assertEqual(gnc.mock_calls, [mock.call("43")])
        self.assertEqual(cn.mock_calls, [])
        rn.reset_mock()
        gnc.reset_mock()

        # 1 TAB -> 1-, 12- (and refresh_nameplates)
        gnc.configure_mock(return_value=["1-", "12-"])
        matches = yield deferToThread(cabc, "1")
        self.assertEqual(matches, ["1-", "12-"])
        self.assertEqual(rn.mock_calls, [mock.call()])
        self.assertEqual(gnc.mock_calls, [mock.call("1")])
        self.assertEqual(cn.mock_calls, [])
        rn.reset_mock()
        gnc.reset_mock()

        # 12 TAB -> 12- (and refresh_nameplates)
        # I wouldn't mind if it didn't refresh the nameplates here, but meh
        gnc.configure_mock(return_value=["12-"])
        matches = yield deferToThread(cabc, "12")
        self.assertEqual(matches, ["12-"])
        self.assertEqual(rn.mock_calls, [mock.call()])
        self.assertEqual(gnc.mock_calls, [mock.call("12")])
        self.assertEqual(cn.mock_calls, [])
        rn.reset_mock()
        gnc.reset_mock()

        # 12- TAB -> 12- {all words} (claim, no refresh)
        gnc.configure_mock(return_value=["12-"])
        gwc.configure_mock(return_value=["and-", "ark-", "aaah!zombies!!-"])
        matches = yield deferToThread(cabc, "12-")
        self.assertEqual(matches, ["12-aaah!zombies!!-", "12-and-", "12-ark-"])
        self.assertEqual(rn.mock_calls, [])
        self.assertEqual(gnc.mock_calls, [])
        self.assertEqual(cn.mock_calls, [mock.call("12")])
        self.assertEqual(gwc.mock_calls, [mock.call("")])
        cn.reset_mock()
        gwc.reset_mock()

        # TODO: another path with "3 TAB" then "34-an TAB", so the claim
        # happens in the second call (and it waits for the wordlist)

        # 12-a TAB -> 12-and- 12-ark- 12-aaah!zombies!!-
        gnc.configure_mock(side_effect=ValueError)
        gwc.configure_mock(return_value=["and-", "ark-", "aaah!zombies!!-"])
        matches = yield deferToThread(cabc, "12-a")
        # matches are always sorted
        self.assertEqual(matches, ["12-aaah!zombies!!-", "12-and-", "12-ark-"])
        self.assertEqual(rn.mock_calls, [])
        self.assertEqual(gnc.mock_calls, [])
        self.assertEqual(cn.mock_calls, [])
        self.assertEqual(gwc.mock_calls, [mock.call("a")])
        gwc.reset_mock()

        # 12-and-b TAB -> 12-and-bat 12-and-bet 12-and-but
        gnc.configure_mock(side_effect=ValueError)
        # wordlist knows the code length, so doesn't add hyphens to these
        gwc.configure_mock(return_value=["and-bat", "and-bet", "and-but"])
        matches = yield deferToThread(cabc, "12-and-b")
        self.assertEqual(matches, ["12-and-bat", "12-and-bet", "12-and-but"])
        self.assertEqual(rn.mock_calls, [])
        self.assertEqual(gnc.mock_calls, [])
        self.assertEqual(cn.mock_calls, [])
        self.assertEqual(gwc.mock_calls, [mock.call("and-b")])
        gwc.reset_mock()

        yield deferToThread(c.finish, "12-and-bat")
        self.assertEqual(cw.mock_calls, [mock.call("and-bat")])

    def test_incomplete_code(self):
        helper = mock.Mock()
        c = CodeInputter(helper, "reactor")
        c.bcft = fake_blockingCallFromThread
        with self.assertRaises(KeyFormatError) as e:
            c.finish("1")
        self.assertEqual(str(e.exception), "incomplete wormhole code")

    @inlineCallbacks
    def test_rollback_nameplate_during_completion(self):
        helper = mock.Mock()
        gwc = helper.get_word_completions = mock.Mock()
        gwc.configure_mock(return_value=["code", "court"])
        c = CodeInputter(helper, reactor)
        cabc = c._commit_and_build_completions
        matches = yield deferToThread(cabc, "1-co")  # this commits us to 1-
        self.assertEqual(helper.mock_calls, [
            mock.call.choose_nameplate("1"),
            mock.call.when_wordlist_is_available(),
            mock.call.get_word_completions("co")
        ])
        self.assertEqual(matches, ["1-code", "1-court"])
        helper.reset_mock()
        with self.assertRaises(AlreadyInputNameplateError) as e:
            yield deferToThread(cabc, "2-co")
        self.assertEqual(
            str(e.exception), "nameplate (1-) already entered, cannot go back")
        self.assertEqual(helper.mock_calls, [])

    @inlineCallbacks
    def test_rollback_nameplate_during_finish(self):
        helper = mock.Mock()
        gwc = helper.get_word_completions = mock.Mock()
        gwc.configure_mock(return_value=["code", "court"])
        c = CodeInputter(helper, reactor)
        cabc = c._commit_and_build_completions
        matches = yield deferToThread(cabc, "1-co")  # this commits us to 1-
        self.assertEqual(helper.mock_calls, [
            mock.call.choose_nameplate("1"),
            mock.call.when_wordlist_is_available(),
            mock.call.get_word_completions("co")
        ])
        self.assertEqual(matches, ["1-code", "1-court"])
        helper.reset_mock()
        with self.assertRaises(AlreadyInputNameplateError) as e:
            yield deferToThread(c.finish, "2-code")
        self.assertEqual(
            str(e.exception), "nameplate (1-) already entered, cannot go back")
        self.assertEqual(helper.mock_calls, [])

    @mock.patch("wormhole._rlcompleter.stderr")
    def test_warn_readline(self, stderr):
        # there is no good way to test that this function gets used at the
        # right time, since it involves a reactor and a "system event
        # trigger", but let's at least make sure it's invocable
        warn_readline()
        expected = "\nCommand interrupted: please press Return to quit"
        self.assertEqual(stderr.mock_calls,
                         [mock.call.write(expected),
                          mock.call.write("\n")])



import os
import sys

from twisted.trial import unittest

from unittest import mock

from ..cli.public_relay import RENDEZVOUS_RELAY, TRANSIT_RELAY
from .common import config


class Send(unittest.TestCase):
    def test_baseline(self):
        cfg = config("send", "--text", "hi")
        self.assertEqual(cfg.what, None)
        self.assertEqual(cfg.code, None)
        self.assertEqual(cfg.code_length, 2)
        self.assertEqual(cfg.dump_timing, None)
        self.assertEqual(cfg.hide_progress, False)
        self.assertEqual(cfg.listen, True)
        self.assertEqual(cfg.appid, None)
        self.assertEqual(cfg.relay_url, RENDEZVOUS_RELAY)
        self.assertEqual(cfg.transit_helper, TRANSIT_RELAY)
        self.assertEqual(cfg.text, "hi")
        self.assertEqual(cfg.tor, False)
        self.assertEqual(cfg.verify, False)
        self.assertEqual(cfg.zeromode, False)

    def test_appid(self):
        cfg = config("--appid", "xyz", "send", "--text", "hi")
        self.assertEqual(cfg.appid, "xyz")
        cfg = config("--appid=xyz", "send", "--text", "hi")
        self.assertEqual(cfg.appid, "xyz")

    def test_file(self):
        cfg = config("send", "fn")
        self.assertEqual(cfg.what, u"fn")
        self.assertEqual(cfg.text, None)

    def test_text(self):
        cfg = config("send", "--text", "hi")
        self.assertEqual(cfg.what, None)
        self.assertEqual(cfg.text, u"hi")

    def test_nolisten(self):
        cfg = config("send", "--no-listen", "fn")
        self.assertEqual(cfg.listen, False)

    def test_code(self):
        cfg = config("send", "--code", "1-abc", "fn")
        self.assertEqual(cfg.code, u"1-abc")

    def test_code_length(self):
        cfg = config("send", "-c", "3", "fn")
        self.assertEqual(cfg.code_length, 3)

    def test_dump_timing(self):
        cfg = config("--dump-timing", "tx.json", "send", "fn")
        self.assertEqual(cfg.dump_timing, "tx.json")

    def test_hide_progress(self):
        cfg = config("send", "--hide-progress", "fn")
        self.assertEqual(cfg.hide_progress, True)

    def test_tor(self):
        cfg = config("send", "--tor", "fn")
        self.assertEqual(cfg.tor, True)

    def test_verify(self):
        cfg = config("send", "--verify", "fn")
        self.assertEqual(cfg.verify, True)

    def test_zeromode(self):
        cfg = config("send", "-0", "fn")
        self.assertEqual(cfg.zeromode, True)

    def test_relay_env_var(self):
        relay_url = str(mock.sentinel.relay_url)
        with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
            cfg = config("send")
        self.assertEqual(cfg.relay_url, relay_url)

        # Make sure cmd line option overrides environment variable
        relay_url_2 = str(mock.sentinel.relay_url_2)
        with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
            cfg = config("--relay-url", relay_url_2, "send")
        self.assertEqual(cfg.relay_url, relay_url_2)

    def test_transit_env_var(self):
        transit_url = str(mock.sentinel.transit_url)
        with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
            cfg = config("send")
        self.assertEqual(cfg.transit_helper, transit_url)

        # Make sure cmd line option overrides environment variable
        transit_url_2 = str(mock.sentinel.transit_url_2)
        with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
            cfg = config("--transit-helper", transit_url_2, "send")
        self.assertEqual(cfg.transit_helper, transit_url_2)


class Receive(unittest.TestCase):
    def test_baseline(self):
        cfg = config("receive")
        self.assertEqual(cfg.accept_file, False)
        self.assertEqual(cfg.code, None)
        self.assertEqual(cfg.code_length, 2)
        self.assertEqual(cfg.dump_timing, None)
        self.assertEqual(cfg.hide_progress, False)
        self.assertEqual(cfg.listen, True)
        self.assertEqual(cfg.only_text, False)
        self.assertEqual(cfg.output_file, None)
        self.assertEqual(cfg.appid, None)
        self.assertEqual(cfg.relay_url, RENDEZVOUS_RELAY)
        self.assertEqual(cfg.transit_helper, TRANSIT_RELAY)
        self.assertEqual(cfg.tor, False)
        self.assertEqual(cfg.verify, False)
        self.assertEqual(cfg.zeromode, False)

    def test_appid(self):
        cfg = config("--appid", "xyz", "receive")
        self.assertEqual(cfg.appid, "xyz")
        cfg = config("--appid=xyz", "receive")
        self.assertEqual(cfg.appid, "xyz")

    def test_nolisten(self):
        cfg = config("receive", "--no-listen")
        self.assertEqual(cfg.listen, False)

    def test_code(self):
        cfg = config("receive", "1-abc")
        self.assertEqual(cfg.code, u"1-abc")

    def test_code_length(self):
        cfg = config("receive", "-c", "3", "--allocate")
        self.assertEqual(cfg.code_length, 3)

    def test_dump_timing(self):
        cfg = config("--dump-timing", "tx.json", "receive")
        self.assertEqual(cfg.dump_timing, "tx.json")

    def test_hide_progress(self):
        cfg = config("receive", "--hide-progress")
        self.assertEqual(cfg.hide_progress, True)

    def test_tor(self):
        cfg = config("receive", "--tor")
        self.assertEqual(cfg.tor, True)

    def test_verify(self):
        cfg = config("receive", "--verify")
        self.assertEqual(cfg.verify, True)

    def test_zeromode(self):
        cfg = config("receive", "-0")
        self.assertEqual(cfg.zeromode, True)

    def test_only_text(self):
        cfg = config("receive", "-t")
        self.assertEqual(cfg.only_text, True)

    def test_accept_file(self):
        cfg = config("receive", "--accept-file")
        self.assertEqual(cfg.accept_file, True)

    def test_output_file(self):
        cfg = config("receive", "--output-file", "fn")
        self.assertEqual(cfg.output_file, u"fn")

    def test_relay_env_var(self):
        relay_url = str(mock.sentinel.relay_url)
        with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
            cfg = config("receive")
        self.assertEqual(cfg.relay_url, relay_url)

        # Make sure cmd line option overrides environment variable
        relay_url_2 = str(mock.sentinel.relay_url_2)
        with mock.patch.dict(os.environ, WORMHOLE_RELAY_URL=relay_url):
            cfg = config("--relay-url", relay_url_2, "receive")
        self.assertEqual(cfg.relay_url, relay_url_2)

    def test_transit_env_var(self):
        transit_url = str(mock.sentinel.transit_url)
        with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
            cfg = config("receive")
        self.assertEqual(cfg.transit_helper, transit_url)

        # Make sure cmd line option overrides environment variable
        transit_url_2 = str(mock.sentinel.transit_url_2)
        with mock.patch.dict(os.environ, WORMHOLE_TRANSIT_HELPER=transit_url):
            cfg = config("--transit-helper", transit_url_2, "receive")
        self.assertEqual(cfg.transit_helper, transit_url_2)


class Config(unittest.TestCase):
    def test_send(self):
        cfg = config("send")
        self.assertEqual(cfg.stdout, sys.stdout)

    def test_receive(self):
        cfg = config("receive")
        self.assertEqual(cfg.stdout, sys.stdout)



import gc
import io
from binascii import hexlify, unhexlify

from nacl.exceptions import CryptoError
from nacl.secret import SecretBox
from twisted.internet import address, defer, endpoints, error, protocol, task
from twisted.internet.defer import gatherResults, inlineCallbacks
from twisted.python import log
from twisted.test import proto_helpers
from twisted.trial import unittest

from unittest import mock
from wormhole_transit_relay import transit_server

from .. import transit
from .._hints import DirectTCPV1Hint
from ..errors import InternalError
from ..util import HKDF
from .common import ServerBase


class Highlander(unittest.TestCase):
    def test_one_winner(self):
        cancelled = set()
        contenders = [
            defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(5)
        ]
        d = transit.there_can_be_only_one(contenders)
        self.assertNoResult(d)
        contenders[0].errback(ValueError())
        self.assertNoResult(d)
        contenders[1].errback(TypeError())
        self.assertNoResult(d)
        contenders[2].callback("yay")
        self.assertEqual(self.successResultOf(d), "yay")
        self.assertEqual(cancelled, set([3, 4]))

    def test_there_might_also_be_none(self):
        cancelled = set()
        contenders = [
            defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
        ]
        d = transit.there_can_be_only_one(contenders)
        self.assertNoResult(d)
        contenders[0].errback(ValueError())
        self.assertNoResult(d)
        contenders[1].errback(TypeError())
        self.assertNoResult(d)
        contenders[2].errback(TypeError())
        self.assertNoResult(d)
        contenders[3].errback(NameError())
        self.failureResultOf(d, ValueError)  # first failure is recorded
        self.assertEqual(cancelled, set())

    def test_cancel_early(self):
        cancelled = set()
        contenders = [
            defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
        ]
        d = transit.there_can_be_only_one(contenders)
        self.assertNoResult(d)
        self.assertEqual(cancelled, set())
        d.cancel()
        self.failureResultOf(d, defer.CancelledError)
        self.assertEqual(cancelled, set(range(4)))

    def test_cancel_after_one_failure(self):
        cancelled = set()
        contenders = [
            defer.Deferred(lambda d, i=i: cancelled.add(i)) for i in range(4)
        ]
        d = transit.there_can_be_only_one(contenders)
        self.assertNoResult(d)
        self.assertEqual(cancelled, set())
        contenders[0].errback(ValueError())
        d.cancel()
        self.failureResultOf(d, ValueError)
        self.assertEqual(cancelled, set([1, 2, 3]))


class Forever(unittest.TestCase):
    def _forever_setup(self):
        clock = task.Clock()
        c = transit.Common("", reactor=clock)
        cancelled = []
        d0 = defer.Deferred(cancelled.append)
        d = c._not_forever(1.0, d0)
        return c, clock, d0, d, cancelled

    def test_not_forever_fires(self):
        c, clock, d0, d, cancelled = self._forever_setup()
        self.assertNoResult(d)
        self.assertEqual(cancelled, [])
        d.callback(1)
        self.assertEqual(self.successResultOf(d), 1)
        self.assertEqual(cancelled, [])
        self.assertNot(clock.getDelayedCalls())

    def test_not_forever_errs(self):
        c, clock, d0, d, cancelled = self._forever_setup()
        self.assertNoResult(d)
        self.assertEqual(cancelled, [])
        d.errback(ValueError())
        self.assertEqual(cancelled, [])
        self.failureResultOf(d, ValueError)
        self.assertNot(clock.getDelayedCalls())

    def test_not_forever_cancel_early(self):
        c, clock, d0, d, cancelled = self._forever_setup()
        self.assertNoResult(d)
        self.assertEqual(cancelled, [])
        d.cancel()
        self.assertEqual(cancelled, [d0])
        self.failureResultOf(d, defer.CancelledError)
        self.assertNot(clock.getDelayedCalls())

    def test_not_forever_timeout(self):
        c, clock, d0, d, cancelled = self._forever_setup()
        self.assertNoResult(d)
        self.assertEqual(cancelled, [])
        clock.advance(2.0)
        self.assertEqual(cancelled, [d0])
        self.failureResultOf(d, defer.CancelledError)
        self.assertNot(clock.getDelayedCalls())


class Misc(unittest.TestCase):
    def test_allocate_port(self):
        portno = transit.allocate_tcp_port()
        self.assertIsInstance(portno, int)

    def test_allocate_port_no_reuseaddr(self):
        mock_sys = mock.Mock()
        mock_sys.platform = "cygwin"
        with mock.patch("wormhole.transit.sys", mock_sys):
            portno = transit.allocate_tcp_port()
        self.assertIsInstance(portno, int)


LOOPADDR = "127.0.0.1"
OTHERADDR = "1.2.3.4"


class Basic(unittest.TestCase):
    @inlineCallbacks
    def test_relay_hints(self):
        URL = "tcp:host:1234"
        c = transit.Common(URL, no_listen=True)
        hints = yield c.get_connection_hints()
        self.assertEqual(hints, [{
            "type":
            "relay-v1",
            "hints": [{
                "type": "direct-tcp-v1",
                "hostname": "host",
                "port": 1234,
                "priority": 0.0
            }],
        }])
        self.assertRaises(InternalError, transit.Common, 123)

    @inlineCallbacks
    def test_no_relay_hints(self):
        c = transit.Common(None, no_listen=True)
        hints = yield c.get_connection_hints()
        self.assertEqual(hints, [])

    def test_ignore_bad_hints(self):
        c = transit.Common("")
        c.add_connection_hints([{"type": "unknown"}])
        c.add_connection_hints([{
            "type": "relay-v1",
            "hints": [{
                "type": "unknown"
            }]
        }])
        self.assertEqual(c._their_direct_hints, [])
        self.assertEqual(c._our_relay_hints, set())

    def test_ignore_localhost_hint_orig(self):
        # this actually starts the listener
        c = transit.TransitSender("")
        hints = self.successResultOf(c.get_connection_hints())
        c._stop_listening()
        # If there are non-localhost hints, then localhost hints should be
        # removed. But if the only hint is localhost, it should stay.
        if len(hints) == 1:
            if hints[0]["hostname"] == "127.0.0.1":
                return
        for hint in hints:
            self.assertFalse(hint["hostname"] == "127.0.0.1")

    def test_ignore_localhost_hint(self):
        # this actually starts the listener
        c = transit.TransitSender("")
        with mock.patch(
                "wormhole.ipaddrs.find_addresses",
                return_value=[LOOPADDR, OTHERADDR]):
            hints = self.successResultOf(c.get_connection_hints())
        c._stop_listening()
        # If there are non-localhost hints, then localhost hints should be
        # removed.
        self.assertEqual(len(hints), 1)
        self.assertEqual(hints[0]["hostname"], "1.2.3.4")

    def test_keep_only_localhost_hint(self):
        # this actually starts the listener
        c = transit.TransitSender("")
        with mock.patch(
                "wormhole.ipaddrs.find_addresses", return_value=[LOOPADDR]):
            hints = self.successResultOf(c.get_connection_hints())
        c._stop_listening()
        # If the only hint is localhost, it should stay.
        self.assertEqual(len(hints), 1)
        self.assertEqual(hints[0]["hostname"], "127.0.0.1")

    def test_abilities(self):
        c = transit.Common(None, no_listen=True)
        abilities = c.get_connection_abilities()
        self.assertEqual(abilities, [
            {
                "type": "direct-tcp-v1"
            },
            {
                "type": "relay-v1"
            },
        ])

    def test_transit_key_wait(self):
        KEY = b"123"
        c = transit.Common("")
        d = c._get_transit_key()
        self.assertNoResult(d)
        c.set_transit_key(KEY)
        self.assertEqual(self.successResultOf(d), KEY)

    def test_transit_key_already_set(self):
        KEY = b"123"
        c = transit.Common("")
        c.set_transit_key(KEY)
        d = c._get_transit_key()
        self.assertEqual(self.successResultOf(d), KEY)

    def test_transit_keys(self):
        KEY = b"123"
        s = transit.TransitSender("")
        s.set_transit_key(KEY)
        r = transit.TransitReceiver("")
        r.set_transit_key(KEY)

        self.assertEqual(s._send_this(), (
            b"transit sender "
            b"559bdeae4b49fa6a23378d2b68f4c7e69378615d4af049c371c6a26e82391089"
            b" ready\n\n"))
        self.assertEqual(s._send_this(), r._expect_this())

        self.assertEqual(r._send_this(), (
            b"transit receiver "
            b"ed447528194bac4c00d0c854b12a97ce51413d89aa74d6304475f516fdc23a1b"
            b" ready\n\n"))
        self.assertEqual(r._send_this(), s._expect_this())

        self.assertEqual(
            hexlify(s._sender_record_key()),
            b"5a2fba3a9e524ab2e2823ff53b05f946896f6e4ce4e282ffd8e3ac0e5e9e0cda"
        )
        self.assertEqual(
            hexlify(s._sender_record_key()), hexlify(r._receiver_record_key()))

        self.assertEqual(
            hexlify(r._sender_record_key()),
            b"eedb143117249f45b39da324decf6bd9aae33b7ccd58487436de611a3c6b871d"
        )
        self.assertEqual(
            hexlify(r._sender_record_key()), hexlify(s._receiver_record_key()))

    def test_connection_ready(self):
        s = transit.TransitSender("")
        self.assertEqual(s.connection_ready("p1"), "go")
        self.assertEqual(s._winner, "p1")
        self.assertEqual(s.connection_ready("p2"), "nevermind")
        self.assertEqual(s._winner, "p1")

        r = transit.TransitReceiver("")
        self.assertEqual(r.connection_ready("p1"), "wait-for-decision")
        self.assertEqual(r.connection_ready("p2"), "wait-for-decision")


class Listener(unittest.TestCase):
    def test_listener(self):
        c = transit.Common("")
        hints, ep = c._build_listener()
        self.assertIsInstance(hints, (list, set))
        if hints:
            self.assertIsInstance(hints[0], DirectTCPV1Hint)
        self.assertIsInstance(ep, endpoints.TCP4ServerEndpoint)

    def test_get_direct_hints(self):
        # this actually starts the listener
        c = transit.TransitSender("")

        d = c.get_connection_hints()
        hints = self.successResultOf(d)

        # the hints are supposed to be cached, so calling this twice won't
        # start a second listener
        self.assert_(c._listener)
        d2 = c.get_connection_hints()
        self.assertEqual(self.successResultOf(d2), hints)

        c._stop_listening()


class DummyProtocol(protocol.Protocol):
    def __init__(self):
        self.buf = b""
        self._count = None
        self._d2 = None

    def wait_for(self, count):
        if len(self.buf) >= count:
            data = self.buf[:count]
            self.buf = self.buf[count:]
            return defer.succeed(data)
        self._d = defer.Deferred()
        self._count = count
        return self._d

    def dataReceived(self, data):
        self.buf += data
        # print("oDR", self._count, len(self.buf))
        if self._count is not None and len(self.buf) >= self._count:
            got = self.buf[:self._count]
            self.buf = self.buf[self._count:]
            self._count = None
            self._d.callback(got)

    def wait_for_disconnect(self):
        self._d2 = defer.Deferred()
        return self._d2

    def connectionLost(self, reason):
        if self._d2:
            self._d2.callback(None)


class FakeTransport:
    signalConnectionLost = True

    def __init__(self, p, peeraddr):
        self.protocol = p
        self._peeraddr = peeraddr
        self._buf = b""
        self._connected = True

    def write(self, data):
        self._buf += data

    def loseConnection(self):
        self._connected = False
        if self.signalConnectionLost:
            self.protocol.connectionLost()

    def getPeer(self):
        return self._peeraddr

    def read_buf(self):
        b = self._buf
        self._buf = b""
        return b


class RandomError(Exception):
    pass


class MockConnection:
    def __init__(self, owner, relay_handshake, start, description):
        self.owner = owner
        self.relay_handshake = relay_handshake
        self.start = start
        self._description = description

        def cancel(d):
            self._cancelled = True

        self._d = defer.Deferred(cancel)
        self._start_negotiation_called = False
        self._cancelled = False

    def startNegotiation(self):
        self._start_negotiation_called = True
        return self._d


class InboundConnectionFactory(unittest.TestCase):
    def test_describe(self):
        f = transit.InboundConnectionFactory(None)
        addrH = address.HostnameAddress("example.com", 1234)
        self.assertEqual(f._describePeer(addrH), "<-example.com:1234")
        addr4 = address.IPv4Address("TCP", "1.2.3.4", 1234)
        self.assertEqual(f._describePeer(addr4), "<-1.2.3.4:1234")
        addr6 = address.IPv6Address("TCP", "::1", 1234)
        self.assertEqual(f._describePeer(addr6), "<-::1:1234")
        addrU = address.UNIXAddress("/dev/unlikely")
        self.assertEqual(
            f._describePeer(addrU), "<-UNIXAddress('/dev/unlikely')")

    def test_success(self):
        f = transit.InboundConnectionFactory("owner")
        f.protocol = MockConnection
        d = f.whenDone()
        self.assertNoResult(d)

        addr = address.HostnameAddress("example.com", 1234)
        p = f.buildProtocol(addr)
        self.assertIsInstance(p, MockConnection)
        self.assertEqual(p.owner, "owner")
        self.assertEqual(p.relay_handshake, None)
        self.assertEqual(p._start_negotiation_called, False)
        # meh .start

        # this is normally called from Connection.connectionMade
        f.connectionWasMade(p)
        self.assertEqual(p._start_negotiation_called, True)
        self.assertNoResult(d)
        self.assertEqual(p._description, "<-example.com:1234")

        p._d.callback(p)
        self.assertEqual(self.successResultOf(d), p)

    def test_one_fail_one_success(self):
        f = transit.InboundConnectionFactory("owner")
        f.protocol = MockConnection
        d = f.whenDone()
        self.assertNoResult(d)

        addr1 = address.HostnameAddress("example.com", 1234)
        addr2 = address.HostnameAddress("example.com", 5678)
        p1 = f.buildProtocol(addr1)
        p2 = f.buildProtocol(addr2)

        f.connectionWasMade(p1)
        f.connectionWasMade(p2)
        self.assertNoResult(d)

        p1._d.errback(transit.BadHandshake("nope"))
        self.assertNoResult(d)
        p2._d.callback(p2)
        self.assertEqual(self.successResultOf(d), p2)

    def test_first_success_wins(self):
        f = transit.InboundConnectionFactory("owner")
        f.protocol = MockConnection
        d = f.whenDone()
        self.assertNoResult(d)

        addr1 = address.HostnameAddress("example.com", 1234)
        addr2 = address.HostnameAddress("example.com", 5678)
        p1 = f.buildProtocol(addr1)
        p2 = f.buildProtocol(addr2)

        f.connectionWasMade(p1)
        f.connectionWasMade(p2)
        self.assertNoResult(d)

        p1._d.callback(p1)
        self.assertEqual(self.successResultOf(d), p1)
        self.assertEqual(p1._cancelled, False)
        self.assertEqual(p2._cancelled, True)

    def test_log_other_errors(self):
        f = transit.InboundConnectionFactory("owner")
        f.protocol = MockConnection
        d = f.whenDone()
        self.assertNoResult(d)

        addr = address.HostnameAddress("example.com", 1234)
        p1 = f.buildProtocol(addr)

        # if the Connection protocol throws an unexpected error, that should
        # get logged to the Twisted logs (as an Unhandled Error in Deferred)
        # so we can diagnose the bug
        f.connectionWasMade(p1)
        our_error = RandomError("boom1")
        p1._d.errback(our_error)
        self.assertNoResult(d)

        log.msg("=== note: the next RandomError is expected ===")
        # Make sure the Deferred has gone out of scope, so the UnhandledError
        # happens quickly. We must manually break the gc cycle.
        del p1._d
        gc.collect()  # make PyPy happy
        errors = self.flushLoggedErrors(RandomError)
        self.assertEqual(1, len(errors))
        self.assertEqual(our_error, errors[0].value)
        log.msg("=== note: the preceding RandomError was expected ===")

    def test_cancel(self):
        f = transit.InboundConnectionFactory("owner")
        f.protocol = MockConnection
        d = f.whenDone()
        self.assertNoResult(d)

        addr1 = address.HostnameAddress("example.com", 1234)
        addr2 = address.HostnameAddress("example.com", 5678)
        p1 = f.buildProtocol(addr1)
        p2 = f.buildProtocol(addr2)

        f.connectionWasMade(p1)
        f.connectionWasMade(p2)
        self.assertNoResult(d)

        d.cancel()

        self.failureResultOf(d, defer.CancelledError)
        self.assertEqual(p1._cancelled, True)
        self.assertEqual(p2._cancelled, True)


# XXX check descriptions


class OutboundConnectionFactory(unittest.TestCase):
    def test_success(self):
        f = transit.OutboundConnectionFactory("owner", "relay_handshake",
                                              "description")
        f.protocol = MockConnection

        addr = address.HostnameAddress("example.com", 1234)
        p = f.buildProtocol(addr)
        self.assertIsInstance(p, MockConnection)
        self.assertEqual(p.owner, "owner")
        self.assertEqual(p.relay_handshake, "relay_handshake")
        self.assertEqual(p._start_negotiation_called, False)
        # meh .start

        # this is normally called from Connection.connectionMade
        f.connectionWasMade(p)  # no-op for outbound
        self.assertEqual(p._start_negotiation_called, False)


class MockOwner:
    _connection_ready_called = False

    def connection_ready(self, connection):
        self._connection_ready_called = True
        self._connection = connection
        return self._state

    def _send_this(self):
        return b"send_this"

    def _expect_this(self):
        return b"expect_this"

    def _sender_record_key(self):
        return b"s" * 32

    def _receiver_record_key(self):
        return b"r" * 32


class MockFactory:
    _connectionWasMade_called = False

    def connectionWasMade(self, p):
        self._connectionWasMade_called = True
        self._p = p


class Connection(unittest.TestCase):
    # exercise the Connection protocol class

    def test_check_and_remove(self):
        c = transit.Connection(None, None, None, "description")
        c.buf = b""
        EXP = b"expectation"
        self.assertFalse(c._check_and_remove(EXP))
        self.assertEqual(c.buf, b"")

        c.buf = b"unexpected"
        e = self.assertRaises(transit.BadHandshake, c._check_and_remove, EXP)
        self.assertEqual(
            str(e), "got %r want %r" % (b'unexpected', b'expectation'))
        self.assertEqual(c.buf, b"unexpected")

        c.buf = b"expect"
        self.assertFalse(c._check_and_remove(EXP))
        self.assertEqual(c.buf, b"expect")

        c.buf = b"expectation"
        self.assertTrue(c._check_and_remove(EXP))
        self.assertEqual(c.buf, b"")

        c.buf = b"expectation exceeded"
        self.assertTrue(c._check_and_remove(EXP))
        self.assertEqual(c.buf, b" exceeded")

    def test_describe(self):
        c = transit.Connection(None, None, None, "description")
        self.assertEqual(c.describe(), "description")

    def test_sender_accepting(self):
        relay_handshake = None
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, relay_handshake, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        owner._state = "go"
        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(t.read_buf(), b"go\n")
        self.assertEqual(t._connected, True)
        self.assertEqual(c.state, "records")
        self.assertEqual(self.successResultOf(d), c)

        c.close()
        self.assertEqual(t._connected, False)

    def test_sender_rejecting(self):
        relay_handshake = None
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, relay_handshake, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        owner._state = "nevermind"
        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(t.read_buf(), b"nevermind\n")
        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")
        f = self.failureResultOf(d, transit.BadHandshake)
        self.assertEqual(str(f.value), "abandoned")

    def test_handshake_other_error(self):
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)
        c.state = RandomError("boom2")
        self.assertRaises(RandomError, c.dataReceived, b"surprise!")
        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")
        self.failureResultOf(d, RandomError)

    def test_handshake_bad_state(self):
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)
        c.state = "unknown-bogus-state"
        self.assertRaises(ValueError, c.dataReceived, b"surprise!")
        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")
        self.failureResultOf(d, ValueError)

    def test_relay_handshake(self):
        relay_handshake = b"relay handshake"
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, relay_handshake, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)
        self.assertEqual(t.read_buf(), b"")  # quiet until startNegotiation

        owner._state = "go"
        d = c.startNegotiation()
        self.assertEqual(t.read_buf(), relay_handshake)
        self.assertEqual(c.state, "relay")  # waiting for OK from relay

        c.dataReceived(b"ok\n")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertEqual(c.state, "handshake")

        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(c.state, "records")
        self.assertEqual(self.successResultOf(d), c)

        self.assertEqual(t.read_buf(), b"go\n")

    def test_relay_handshake_bad(self):
        relay_handshake = b"relay handshake"
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, relay_handshake, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)
        self.assertEqual(t.read_buf(), b"")  # quiet until startNegotiation

        owner._state = "go"
        d = c.startNegotiation()
        self.assertEqual(t.read_buf(), relay_handshake)
        self.assertEqual(c.state, "relay")  # waiting for OK from relay

        c.dataReceived(b"not ok\n")
        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")

        f = self.failureResultOf(d, transit.BadHandshake)
        self.assertEqual(
            str(f.value), "got %r want %r" % (b"not ok\n", b"ok\n"))

    def test_receiver_accepted(self):
        # we're on the receiving side, so we wait for the sender to decide
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        owner._state = "wait-for-decision"
        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(c.state, "wait-for-decision")
        self.assertNoResult(d)

        c.dataReceived(b"go\n")
        self.assertEqual(c.state, "records")
        self.assertEqual(self.successResultOf(d), c)

    def test_receiver_rejected_politely(self):
        # we're on the receiving side, so we wait for the sender to decide
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        owner._state = "wait-for-decision"
        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(c.state, "wait-for-decision")
        self.assertNoResult(d)

        c.dataReceived(b"nevermind\n")  # polite rejection
        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")
        f = self.failureResultOf(d, transit.BadHandshake)
        self.assertEqual(
            str(f.value), "got %r want %r" % (b"nevermind\n", b"go\n"))

    def test_receiver_rejected_rudely(self):
        # we're on the receiving side, so we wait for the sender to decide
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        self.assertEqual(factory._connectionWasMade_called, True)
        self.assertEqual(factory._p, c)

        owner._state = "wait-for-decision"
        d = c.startNegotiation()
        self.assertEqual(c.state, "handshake")
        self.assertEqual(t.read_buf(), b"send_this")
        self.assertNoResult(d)

        c.dataReceived(b"expect_this")
        self.assertEqual(c.state, "wait-for-decision")
        self.assertNoResult(d)

        t.loseConnection()
        self.assertEqual(t._connected, False)
        f = self.failureResultOf(d, transit.BadHandshake)
        self.assertEqual(str(f.value), "connection lost")

    def test_cancel(self):
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()

        d = c.startNegotiation()
        # while we're waiting for negotiation, we get cancelled
        d.cancel()

        self.assertEqual(t._connected, False)
        self.assertEqual(c.state, "hung up")
        self.failureResultOf(d, defer.CancelledError)

    def test_timeout(self):
        clock = task.Clock()
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")

        def _callLater(period, func):
            clock.callLater(period, func)

        c.callLater = _callLater
        self.assertEqual(c.state, "too-early")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()
        # the timer should now be running
        d = c.startNegotiation()
        # while we're waiting for negotiation, the timer expires
        clock.advance(transit.TIMEOUT + 1.0)

        self.assertEqual(t._connected, False)
        f = self.failureResultOf(d, transit.BadHandshake)
        self.assertEqual(str(f.value), "timeout")

    def make_connection(self):
        owner = MockOwner()
        factory = MockFactory()
        addr = address.HostnameAddress("example.com", 1234)
        c = transit.Connection(owner, None, None, "description")
        t = c.transport = FakeTransport(c, addr)
        c.factory = factory
        c.connectionMade()

        owner._state = "go"
        d = c.startNegotiation()
        c.dataReceived(b"expect_this")
        self.assertEqual(self.successResultOf(d), c)
        t.read_buf()  # flush input buffer, prepare for encrypted records

        return t, c, owner

    def test_records_not_binary(self):
        t, c, owner = self.make_connection()

        RECORD1 = u"not binary"
        with self.assertRaises(InternalError):
            c.send_record(RECORD1)

    def test_records_good(self):
        # now make sure that outbound records are encrypted properly
        t, c, owner = self.make_connection()

        RECORD1 = b"record"
        c.send_record(RECORD1)
        buf = t.read_buf()
        expected = ("%08x" % (24 + len(RECORD1) + 16)).encode("ascii")
        self.assertEqual(hexlify(buf[:4]), expected)
        encrypted = buf[4:]
        receive_box = SecretBox(owner._sender_record_key())
        nonce_buf = encrypted[:SecretBox.NONCE_SIZE]  # assume it's prepended
        nonce = int(hexlify(nonce_buf), 16)
        self.assertEqual(nonce, 0)  # first message gets nonce 0
        decrypted = receive_box.decrypt(encrypted)
        self.assertEqual(decrypted, RECORD1)

        # second message gets nonce 1
        RECORD2 = b"record2"
        c.send_record(RECORD2)
        buf = t.read_buf()
        expected = ("%08x" % (24 + len(RECORD2) + 16)).encode("ascii")
        self.assertEqual(hexlify(buf[:4]), expected)
        encrypted = buf[4:]
        receive_box = SecretBox(owner._sender_record_key())
        nonce_buf = encrypted[:SecretBox.NONCE_SIZE]  # assume it's prepended
        nonce = int(hexlify(nonce_buf), 16)
        self.assertEqual(nonce, 1)
        decrypted = receive_box.decrypt(encrypted)
        self.assertEqual(decrypted, RECORD2)

        # and that we can receive records properly
        inbound_records = []
        c.recordReceived = inbound_records.append
        send_box = SecretBox(owner._receiver_record_key())

        RECORD3 = b"record3"
        nonce_buf = unhexlify("%048x" % 0)  # first nonce must be 0
        encrypted = send_box.encrypt(RECORD3, nonce_buf)
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        c.dataReceived(length[:2])
        c.dataReceived(length[2:])
        c.dataReceived(encrypted[:-2])
        self.assertEqual(inbound_records, [])
        c.dataReceived(encrypted[-2:])
        self.assertEqual(inbound_records, [RECORD3])

        RECORD4 = b"record4"
        nonce_buf = unhexlify("%048x" % 1)  # nonces increment
        encrypted = send_box.encrypt(RECORD4, nonce_buf)
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        c.dataReceived(length[:2])
        c.dataReceived(length[2:])
        c.dataReceived(encrypted[:-2])
        self.assertEqual(inbound_records, [RECORD3])
        c.dataReceived(encrypted[-2:])
        self.assertEqual(inbound_records, [RECORD3, RECORD4])

        # receiving two records at the same time: deliver both
        inbound_records[:] = []
        RECORD5 = b"record5"
        nonce_buf = unhexlify("%048x" % 2)  # nonces increment
        encrypted = send_box.encrypt(RECORD5, nonce_buf)
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        r5 = length + encrypted
        RECORD6 = b"record6"
        nonce_buf = unhexlify("%048x" % 3)  # nonces increment
        encrypted = send_box.encrypt(RECORD6, nonce_buf)
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        r6 = length + encrypted
        c.dataReceived(r5 + r6)
        self.assertEqual(inbound_records, [RECORD5, RECORD6])

    def corrupt(self, orig):
        last_byte = orig[-1:]
        num = int(hexlify(last_byte).decode("ascii"), 16)
        corrupt_num = 256 - num
        as_byte = unhexlify("%02x" % corrupt_num)
        return orig[:-1] + as_byte

    def test_records_corrupt(self):
        # corrupt records should be rejected
        t, c, owner = self.make_connection()

        inbound_records = []
        c.recordReceived = inbound_records.append

        RECORD = b"record"
        send_box = SecretBox(owner._receiver_record_key())
        nonce_buf = unhexlify("%048x" % 0)  # first nonce must be 0
        encrypted = self.corrupt(send_box.encrypt(RECORD, nonce_buf))
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        c.dataReceived(length)
        c.dataReceived(encrypted[:-2])
        self.assertEqual(inbound_records, [])
        self.assertRaises(CryptoError, c.dataReceived, encrypted[-2:])
        self.assertEqual(inbound_records, [])
        # and the connection should have been dropped
        self.assertEqual(t._connected, False)

    def test_out_of_order_nonce(self):
        # an inbound out-of-order nonce should be rejected
        t, c, owner = self.make_connection()

        inbound_records = []
        c.recordReceived = inbound_records.append

        RECORD = b"record"
        send_box = SecretBox(owner._receiver_record_key())
        nonce_buf = unhexlify("%048x" % 1)  # first nonce must be 0
        encrypted = send_box.encrypt(RECORD, nonce_buf)
        length = unhexlify("%08x" % len(encrypted))  # always 4 bytes long
        c.dataReceived(length)
        c.dataReceived(encrypted[:-2])
        self.assertEqual(inbound_records, [])
        self.assertRaises(transit.BadNonce, c.dataReceived, encrypted[-2:])
        self.assertEqual(inbound_records, [])
        # and the connection should have been dropped
        self.assertEqual(t._connected, False)

        # TODO: check that .connectionLost/loseConnection signatures are
        # consistent: zero args, or one arg?

        # XXX: if we don't set the transit key before connecting, what
        # happens? We currently get a type-check assertion from HKDF because
        # the key is None.

    def test_receive_queue(self):
        c = transit.Connection(None, None, None, "description")
        c.transport = FakeTransport(c, None)
        c.transport.signalConnectionLost = False
        c.recordReceived(b"0")
        c.recordReceived(b"1")
        c.recordReceived(b"2")
        d0 = c.receive_record()
        self.assertEqual(self.successResultOf(d0), b"0")
        d1 = c.receive_record()
        d2 = c.receive_record()
        # they must fire in order of receipt, not order of addCallback
        self.assertEqual(self.successResultOf(d2), b"2")
        self.assertEqual(self.successResultOf(d1), b"1")

        d3 = c.receive_record()
        d4 = c.receive_record()
        self.assertNoResult(d3)
        self.assertNoResult(d4)

        c.recordReceived(b"3")
        self.assertEqual(self.successResultOf(d3), b"3")
        self.assertNoResult(d4)

        c.recordReceived(b"4")
        self.assertEqual(self.successResultOf(d4), b"4")

        d5 = c.receive_record()
        c.close()
        self.failureResultOf(d5, error.ConnectionClosed)

    def test_producer(self):
        # a Transit object (receiving data from the remote peer) produces
        # data and writes it into a local Consumer
        c = transit.Connection(None, None, None, "description")
        c.transport = proto_helpers.StringTransport()
        c.recordReceived(b"r1.")
        c.recordReceived(b"r2.")

        consumer = proto_helpers.StringTransport()
        rv = c.connectConsumer(consumer)
        self.assertIs(rv, None)
        self.assertIs(c._consumer, consumer)
        self.assertEqual(consumer.value(), b"r1.r2.")

        self.assertRaises(RuntimeError, c.connectConsumer, consumer)

        c.recordReceived(b"r3.")
        self.assertEqual(consumer.value(), b"r1.r2.r3.")

        c.pauseProducing()
        self.assertEqual(c.transport.producerState, "paused")
        c.resumeProducing()
        self.assertEqual(c.transport.producerState, "producing")

        c.disconnectConsumer()
        self.assertEqual(consumer.producer, None)
        c.connectConsumer(consumer)

        c.stopProducing()
        self.assertEqual(c.transport.producerState, "stopped")

    def test_connectConsumer(self):
        # connectConsumer() takes an optional number of bytes to expect, and
        # fires a Deferred when that many have been written
        c = transit.Connection(None, None, None, "description")
        c._negotiation_d.addErrback(lambda err: None)  # eat it
        c.transport = proto_helpers.StringTransport()
        c.recordReceived(b"r1.")

        consumer = proto_helpers.StringTransport()
        d = c.connectConsumer(consumer, expected=10)
        self.assertEqual(consumer.value(), b"r1.")
        self.assertNoResult(d)

        c.recordReceived(b"r2.")
        self.assertEqual(consumer.value(), b"r1.r2.")
        self.assertNoResult(d)

        c.recordReceived(b"r3.")
        self.assertEqual(consumer.value(), b"r1.r2.r3.")
        self.assertNoResult(d)

        c.recordReceived(b"!")
        self.assertEqual(consumer.value(), b"r1.r2.r3.!")
        self.assertEqual(self.successResultOf(d), 10)

        # that should automatically disconnect the consumer, and subsequent
        # records should get queued, not delivered
        self.assertIs(c._consumer, None)
        c.recordReceived(b"overflow")
        self.assertEqual(consumer.value(), b"r1.r2.r3.!")

        # now test that the Deferred errbacks when the connection is lost
        d = c.connectConsumer(consumer, expected=10)

        c.connectionLost()
        self.failureResultOf(d, error.ConnectionClosed)

    def test_connectConsumer_empty(self):
        # if connectConsumer() expects 0 bytes (e.g. someone is "sending" a
        # zero-length file), make sure it gets woken up right away, so it can
        # disconnect itself, even though no bytes will actually arrive
        c = transit.Connection(None, None, None, "description")
        c._negotiation_d.addErrback(lambda err: None)  # eat it
        c.transport = proto_helpers.StringTransport()

        consumer = proto_helpers.StringTransport()
        d = c.connectConsumer(consumer, expected=0)
        self.assertEqual(self.successResultOf(d), 0)
        self.assertEqual(consumer.value(), b"")
        self.assertIs(c._consumer, None)

    def test_writeToFile(self):
        c = transit.Connection(None, None, None, "description")
        c._negotiation_d.addErrback(lambda err: None)  # eat it
        c.transport = proto_helpers.StringTransport()
        c.recordReceived(b"r1.")

        f = io.BytesIO()
        progress = []
        d = c.writeToFile(f, 10, progress.append)
        self.assertEqual(f.getvalue(), b"r1.")
        self.assertEqual(progress, [3])
        self.assertNoResult(d)

        c.recordReceived(b"r2.")
        self.assertEqual(f.getvalue(), b"r1.r2.")
        self.assertEqual(progress, [3, 3])
        self.assertNoResult(d)

        c.recordReceived(b"r3.")
        self.assertEqual(f.getvalue(), b"r1.r2.r3.")
        self.assertEqual(progress, [3, 3, 3])
        self.assertNoResult(d)

        c.recordReceived(b"!")
        self.assertEqual(f.getvalue(), b"r1.r2.r3.!")
        self.assertEqual(progress, [3, 3, 3, 1])
        self.assertEqual(self.successResultOf(d), 10)

        # that should automatically disconnect the consumer, and subsequent
        # records should get queued, not delivered
        self.assertIs(c._consumer, None)
        c.recordReceived(b"overflow.")
        self.assertEqual(f.getvalue(), b"r1.r2.r3.!")
        self.assertEqual(progress, [3, 3, 3, 1])

        # test what happens when enough data is queued ahead of time
        c.recordReceived(b"second.")  # now "overflow.second."
        c.recordReceived(b"third.")  # now "overflow.second.third."
        f = io.BytesIO()
        d = c.writeToFile(f, 10)
        self.assertEqual(f.getvalue(), b"overflow.second.")  # whole records
        self.assertEqual(self.successResultOf(d), 16)
        self.assertEqual(list(c._inbound_records), [b"third."])

        # now test that the Deferred errbacks when the connection is lost
        d = c.writeToFile(f, 10)

        c.connectionLost()
        self.failureResultOf(d, error.ConnectionClosed)

    def test_consumer(self):
        # a local producer sends data to a consuming Transit object
        c = transit.Connection(None, None, None, "description")
        c.transport = proto_helpers.StringTransport()
        records = []
        c.send_record = records.append

        producer = proto_helpers.StringTransport()
        c.registerProducer(producer, True)
        self.assertIs(c.transport.producer, producer)

        c.write(b"r1.")
        self.assertEqual(records, [b"r1."])

        c.unregisterProducer()
        self.assertEqual(c.transport.producer, None)


class FileConsumer(unittest.TestCase):
    def test_basic(self):
        f = io.BytesIO()
        progress = []
        fc = transit.FileConsumer(f, progress.append)
        self.assertEqual(progress, [])
        self.assertEqual(f.getvalue(), b"")
        fc.write(b"." * 99)
        self.assertEqual(progress, [99])
        self.assertEqual(f.getvalue(), b"." * 99)
        fc.write(b"!")
        self.assertEqual(progress, [99, 1])
        self.assertEqual(f.getvalue(), b"." * 99 + b"!")

    def test_hasher(self):
        hashee = []
        f = io.BytesIO()
        progress = []
        fc = transit.FileConsumer(f, progress.append, hasher=hashee.append)
        self.assertEqual(progress, [])
        self.assertEqual(f.getvalue(), b"")
        self.assertEqual(hashee, [])
        fc.write(b"." * 99)
        self.assertEqual(progress, [99])
        self.assertEqual(f.getvalue(), b"." * 99)
        self.assertEqual(hashee, [b"." * 99])
        fc.write(b"!")
        self.assertEqual(progress, [99, 1])
        self.assertEqual(f.getvalue(), b"." * 99 + b"!")
        self.assertEqual(hashee, [b"." * 99, b"!"])


DIRECT_HINT_JSON = {
    "type": "direct-tcp-v1",
    "hostname": "direct",
    "port": 1234
}
RELAY_HINT_JSON = {
    "type": "relay-v1",
    "hints": [{
        "type": "direct-tcp-v1",
        "hostname": "relay",
        "port": 1234
    }]
}
UNRECOGNIZED_DIRECT_HINT_JSON = {
    "type": "direct-tcp-v1",
    "hostname": ["cannot", "parse", "list"]
}
UNRECOGNIZED_HINT_JSON = {"type": "unknown"}
UNAVAILABLE_HINT_JSON = {
    "type": "direct-tcp-v1",  # e.g. Tor without txtorcon
    "hostname": "unavailable",
    "port": 1234
}
RELAY_HINT2_JSON = {
    "type":
    "relay-v1",
    "hints": [{
        "type": "direct-tcp-v1",
        "hostname": "relay",
        "port": 1234
    }, UNRECOGNIZED_HINT_JSON]
}
UNAVAILABLE_RELAY_HINT_JSON = {
    "type": "relay-v1",
    "hints": [UNAVAILABLE_HINT_JSON]
}


class Transit(unittest.TestCase):
    def setUp(self):
        self._connectors = []
        self._waiters = []
        self._descriptions = []

    def _start_connector(self, ep, description, is_relay=False):
        d = defer.Deferred()
        self._connectors.append(ep)
        self._waiters.append(d)
        self._descriptions.append(description)
        return d

    @inlineCallbacks
    def test_success_direct(self):
        reactor = mock.Mock()
        s = transit.TransitSender("", reactor=reactor)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()  # start the listener
        del hints
        s.add_connection_hints([
            DIRECT_HINT_JSON, UNRECOGNIZED_DIRECT_HINT_JSON,
            UNRECOGNIZED_HINT_JSON
        ])

        s._start_connector = self._start_connector
        d = s.connect()
        self.assertNoResult(d)
        self.assertEqual(len(self._waiters), 1)
        self.assertIsInstance(self._waiters[0], defer.Deferred)

        self._waiters[0].callback("winner")
        self.assertEqual(self.successResultOf(d), "winner")
        self.assertEqual(self._descriptions, ["->tcp:direct:1234"])

    @inlineCallbacks
    def test_success_direct_tor(self):
        clock = task.Clock()
        s = transit.TransitSender("", tor=mock.Mock(), reactor=clock)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()  # start the listener
        del hints
        s.add_connection_hints([DIRECT_HINT_JSON])

        s._start_connector = self._start_connector
        d = s.connect()
        self.assertNoResult(d)
        self.assertEqual(len(self._waiters), 1)
        self.assertIsInstance(self._waiters[0], defer.Deferred)

        self._waiters[0].callback("winner")
        self.assertEqual(self.successResultOf(d), "winner")
        self.assertEqual(self._descriptions, ["tor->tcp:direct:1234"])

    @inlineCallbacks
    def test_success_direct_tor_relay(self):
        clock = task.Clock()
        s = transit.TransitSender("", tor=mock.Mock(), reactor=clock)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()  # start the listener
        del hints
        s.add_connection_hints([RELAY_HINT_JSON])

        s._start_connector = self._start_connector
        d = s.connect()
        # move the clock forward any amount, since relay connections are
        # triggered starting at T+0.0
        clock.advance(1.0)
        self.assertNoResult(d)
        self.assertEqual(len(self._waiters), 1)
        self.assertIsInstance(self._waiters[0], defer.Deferred)

        self._waiters[0].callback("winner")
        self.assertEqual(self.successResultOf(d), "winner")
        self.assertEqual(self._descriptions, ["tor->relay:tcp:relay:1234"])

    def _endpoint_from_hint_obj(self, hint, _tor, _reactor):
        if isinstance(hint, DirectTCPV1Hint):
            if hint.hostname == "unavailable":
                return None
            return hint.hostname
        return None

    @inlineCallbacks
    def test_wait_for_relay(self):
        clock = task.Clock()
        s = transit.TransitSender("", reactor=clock, no_listen=True)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()
        del hints
        s.add_connection_hints(
            [DIRECT_HINT_JSON, UNRECOGNIZED_HINT_JSON, RELAY_HINT_JSON])
        s._start_connector = self._start_connector

        with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                        self._endpoint_from_hint_obj):
            d = s.connect()
            self.assertNoResult(d)
            # the direct connectors are tried right away, but the relay
            # connectors are stalled for a few seconds
            self.assertEqual(self._connectors, ["direct"])

            clock.advance(s.RELAY_DELAY + 1.0)
            self.assertEqual(self._connectors, ["direct", "relay"])

            self._waiters[0].callback("winner")
            self.assertEqual(self.successResultOf(d), "winner")

    @inlineCallbacks
    def test_priorities(self):
        clock = task.Clock()
        s = transit.TransitSender("", reactor=clock, no_listen=True)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()
        del hints
        s.add_connection_hints([
            {
                "type":
                "relay-v1",
                "hints": [{
                    "type": "direct-tcp-v1",
                    "hostname": "relay",
                    "port": 1234
                }]
            },
            {
                "type": "direct-tcp-v1",
                "hostname": "direct",
                "port": 1234
            },
            {
                "type":
                "relay-v1",
                "hints": [{
                    "type": "direct-tcp-v1",
                    "priority": 2.0,
                    "hostname": "relay2",
                    "port": 1234
                }, {
                    "type": "direct-tcp-v1",
                    "priority": 3.0,
                    "hostname": "relay3",
                    "port": 1234
                }]
            },
            {
                "type":
                "relay-v1",
                "hints": [{
                    "type": "direct-tcp-v1",
                    "priority": 2.0,
                    "hostname": "relay4",
                    "port": 1234
                }]
            },
        ])
        s._start_connector = self._start_connector

        with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                        self._endpoint_from_hint_obj):
            d = s.connect()
            self.assertNoResult(d)
            # direct connector should be used first, then the priority=3.0 relay,
            # then the two 2.0 relays, then the (default) 0.0 relay

            self.assertEqual(self._connectors, ["direct"])

            clock.advance(s.RELAY_DELAY + 1.0)
            self.assertEqual(self._connectors, ["direct", "relay3"])

            clock.advance(s.RELAY_DELAY)
            self.assertIn(self._connectors,
                          (["direct", "relay3", "relay2", "relay4"],
                           ["direct", "relay3", "relay4", "relay2"]))

            clock.advance(s.RELAY_DELAY)
            self.assertIn(self._connectors,
                          (["direct", "relay3", "relay2", "relay4", "relay"],
                           ["direct", "relay3", "relay4", "relay2", "relay"]))

            self._waiters[0].callback("winner")
            self.assertEqual(self.successResultOf(d), "winner")

    @inlineCallbacks
    def test_no_direct_hints(self):
        clock = task.Clock()
        s = transit.TransitSender("", reactor=clock, no_listen=True)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()  # start the listener
        del hints
        # include hints that can't be turned into an endpoint at runtime
        s.add_connection_hints([
            UNRECOGNIZED_HINT_JSON, UNAVAILABLE_HINT_JSON, RELAY_HINT2_JSON,
            UNAVAILABLE_RELAY_HINT_JSON
        ])
        s._start_connector = self._start_connector

        with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                        self._endpoint_from_hint_obj):
            d = s.connect()
            self.assertNoResult(d)
            # since there are no usable direct hints, the relay connector will
            # only be stalled for 0 seconds
            self.assertEqual(self._connectors, [])

            clock.advance(0)
            self.assertEqual(self._connectors, ["relay"])

            self._waiters[0].callback("winner")
            self.assertEqual(self.successResultOf(d), "winner")

    @inlineCallbacks
    def test_no_contenders(self):
        clock = task.Clock()
        s = transit.TransitSender("", reactor=clock, no_listen=True)
        s.set_transit_key(b"key")
        hints = yield s.get_connection_hints()  # start the listener
        del hints
        s.add_connection_hints([])  # no hints at all
        s._start_connector = self._start_connector

        with mock.patch("wormhole.transit.endpoint_from_hint_obj",
                        self._endpoint_from_hint_obj):
            d = s.connect()
            f = self.failureResultOf(d, transit.TransitError)
            self.assertEqual(str(f.value), "No contenders for connection")


class RelayHandshake(unittest.TestCase):
    def old_build_relay_handshake(self, key):
        token = HKDF(key, 32, CTXinfo=b"transit_relay_token")
        return (token, b"please relay " + hexlify(token) + b"\n")

    def test_old(self):
        key = b"\x00"
        token, old_handshake = self.old_build_relay_handshake(key)
        tc = transit_server.TransitConnection()
        tc.factory = mock.Mock()
        tc.factory.connection_got_token = mock.Mock()
        tc.dataReceived(old_handshake[:-1])
        self.assertEqual(tc.factory.connection_got_token.mock_calls, [])
        tc.dataReceived(old_handshake[-1:])
        self.assertEqual(tc.factory.connection_got_token.mock_calls,
                         [mock.call(hexlify(token), None, tc)])

    def test_new(self):
        c = transit.Common(None)
        c.set_transit_key(b"\x00")
        new_handshake = c._build_relay_handshake()
        token, old_handshake = self.old_build_relay_handshake(b"\x00")

        tc = transit_server.TransitConnection()
        tc.factory = mock.Mock()
        tc.factory.connection_got_token = mock.Mock()
        tc.dataReceived(new_handshake[:-1])
        self.assertEqual(tc.factory.connection_got_token.mock_calls, [])
        tc.dataReceived(new_handshake[-1:])
        self.assertEqual(
            tc.factory.connection_got_token.mock_calls,
            [mock.call(hexlify(token), c._side.encode("ascii"), tc)])


class Full(ServerBase, unittest.TestCase):
    def doBoth(self, d1, d2):
        return gatherResults([d1, d2], True)

    @inlineCallbacks
    def test_direct(self):
        KEY = b"k" * 32
        s = transit.TransitSender(None)
        r = transit.TransitReceiver(None)

        s.set_transit_key(KEY)
        r.set_transit_key(KEY)

        # TODO: this sometimes fails with EADDRINUSE
        shints = yield s.get_connection_hints()
        rhints = yield r.get_connection_hints()

        s.add_connection_hints(rhints)
        r.add_connection_hints(shints)

        (x, y) = yield self.doBoth(s.connect(), r.connect())
        self.assertIsInstance(x, transit.Connection)
        self.assertIsInstance(y, transit.Connection)

        d = y.receive_record()

        x.send_record(b"record1")
        r = yield d
        self.assertEqual(r, b"record1")

        yield x.close()
        yield y.close()

    @inlineCallbacks
    def test_relay(self):
        KEY = b"k" * 32
        s = transit.TransitSender(self.transit, no_listen=True)
        r = transit.TransitReceiver(self.transit, no_listen=True)

        s.set_transit_key(KEY)
        r.set_transit_key(KEY)

        shints = yield s.get_connection_hints()
        rhints = yield r.get_connection_hints()

        s.add_connection_hints(rhints)
        r.add_connection_hints(shints)

        (x, y) = yield self.doBoth(s.connect(), r.connect())
        self.assertIsInstance(x, transit.Connection)
        self.assertIsInstance(y, transit.Connection)

        d = y.receive_record()

        x.send_record(b"record1")
        r = yield d
        self.assertEqual(r, b"record1")

        yield x.close()
        yield y.close()



from __future__ import print_function, unicode_literals
from unittest import mock
from twisted.trial import unittest
from ..._dilation.connection import (parse_record, encode_record,
                                     KCM, Ping, Pong, Open, Data, Close, Ack)


class Parse(unittest.TestCase):
    def test_parse(self):
        self.assertEqual(parse_record(b"\x00"), KCM())
        self.assertEqual(parse_record(b"\x01\x55\x44\x33\x22"),
                         Ping(ping_id=b"\x55\x44\x33\x22"))
        self.assertEqual(parse_record(b"\x02\x55\x44\x33\x22"),
                         Pong(ping_id=b"\x55\x44\x33\x22"))
        self.assertEqual(parse_record(b"\x03\x00\x00\x02\x01\x00\x00\x01\x00"),
                         Open(scid=513, seqnum=256))
        self.assertEqual(parse_record(b"\x04\x00\x00\x02\x02\x00\x00\x01\x01dataaa"),
                         Data(scid=514, seqnum=257, data=b"dataaa"))
        self.assertEqual(parse_record(b"\x05\x00\x00\x02\x03\x00\x00\x01\x02"),
                         Close(scid=515, seqnum=258))
        self.assertEqual(parse_record(b"\x06\x00\x00\x01\x03"),
                         Ack(resp_seqnum=259))
        with mock.patch("wormhole._dilation.connection.log.err") as le:
            with self.assertRaises(ValueError):
                parse_record(b"\x07unknown")
        self.assertEqual(le.mock_calls,
                         [mock.call("received unknown message type: {}".format(
                             b"\x07unknown"))])

    def test_encode(self):
        self.assertEqual(encode_record(KCM()), b"\x00")
        self.assertEqual(encode_record(Ping(ping_id=b"ping")), b"\x01ping")
        self.assertEqual(encode_record(Pong(ping_id=b"pong")), b"\x02pong")
        self.assertEqual(encode_record(Open(scid=65536, seqnum=16)),
                         b"\x03\x00\x01\x00\x00\x00\x00\x00\x10")
        self.assertEqual(encode_record(Data(scid=65537, seqnum=17, data=b"dataaa")),
                         b"\x04\x00\x01\x00\x01\x00\x00\x00\x11dataaa")
        self.assertEqual(encode_record(Close(scid=65538, seqnum=18)),
                         b"\x05\x00\x01\x00\x02\x00\x00\x00\x12")
        self.assertEqual(encode_record(Ack(resp_seqnum=19)),
                         b"\x06\x00\x00\x00\x13")
        with self.assertRaises(TypeError) as ar:
            encode_record("not a record")
        self.assertEqual(str(ar.exception), "not a record")



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from twisted.internet.interfaces import ITransport
from ..._dilation.connection import _Framer, Frame, Prologue, Disconnect


def make_framer():
    t = mock.Mock()
    alsoProvides(t, ITransport)
    f = _Framer(t, b"outbound_prologue\n", b"inbound_prologue\n")
    return f, t


class Framer(unittest.TestCase):
    def test_bad_prologue_length(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        t.mock_calls[:] = []
        self.assertEqual([], list(f.add_and_parse(b"inbound_")))  # wait for it
        self.assertEqual(t.mock_calls, [])

        with mock.patch("wormhole._dilation.connection.log.msg") as m:
            with self.assertRaises(Disconnect):
                list(f.add_and_parse(b"not the prologue after all"))
        self.assertEqual(m.mock_calls,
                         [mock.call("bad prologue: {}".format(
                             b"inbound_not the p"))])
        self.assertEqual(t.mock_calls, [])

    def test_bad_prologue_newline(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        t.mock_calls[:] = []
        self.assertEqual([], list(f.add_and_parse(b"inbound_")))  # wait for it

        self.assertEqual([], list(f.add_and_parse(b"not")))
        with mock.patch("wormhole._dilation.connection.log.msg") as m:
            with self.assertRaises(Disconnect):
                list(f.add_and_parse(b"\n"))
        self.assertEqual(m.mock_calls,
                         [mock.call("bad prologue: {}".format(
                             b"inbound_not\n"))])
        self.assertEqual(t.mock_calls, [])

    def test_good_prologue(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        t.mock_calls[:] = []
        self.assertEqual([Prologue()],
                         list(f.add_and_parse(b"inbound_prologue\n")))
        self.assertEqual(t.mock_calls, [])

        # now send_frame should work
        f.send_frame(b"frame")
        self.assertEqual(t.mock_calls,
                         [mock.call.write(b"\x00\x00\x00\x05frame")])

    def test_bad_relay(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])
        f.use_relay(b"relay handshake\n")

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"relay handshake\n")])
        t.mock_calls[:] = []
        with mock.patch("wormhole._dilation.connection.log.msg") as m:
            with self.assertRaises(Disconnect):
                list(f.add_and_parse(b"goodbye\n"))
        self.assertEqual(m.mock_calls,
                         [mock.call("bad relay_ok: {}".format(b"goo"))])
        self.assertEqual(t.mock_calls, [])

    def test_good_relay(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])
        f.use_relay(b"relay handshake\n")
        self.assertEqual(t.mock_calls, [])

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"relay handshake\n")])
        t.mock_calls[:] = []

        self.assertEqual([], list(f.add_and_parse(b"ok\n")))
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])

    def test_frame(self):
        f, t = make_framer()
        self.assertEqual(t.mock_calls, [])

        f.connectionMade()
        self.assertEqual(t.mock_calls, [mock.call.write(b"outbound_prologue\n")])
        t.mock_calls[:] = []
        self.assertEqual([Prologue()],
                         list(f.add_and_parse(b"inbound_prologue\n")))
        self.assertEqual(t.mock_calls, [])

        encoded_frame = b"\x00\x00\x00\x05frame"
        self.assertEqual([], list(f.add_and_parse(encoded_frame[:2])))
        self.assertEqual([], list(f.add_and_parse(encoded_frame[2:6])))
        self.assertEqual([Frame(frame=b"frame")],
                         list(f.add_and_parse(encoded_frame[6:])))



# To use the web() option, you should do:
# * cd misc
# * npm install d3@3.5.17 d3-tip@0.6.7 zepto

from __future__ import print_function
import os, sys, time, json, random


streams = sys.argv[1:]
if len(streams) != 2:
    print("run like: python dump-timing.py tx.json rx.json")
    sys.exit(1)
# for now, require sender as first file, receiver as second
# later, allow use of only one file.

data = {}
for i,fn in enumerate(streams):
    name = ["send", "receive"][i]
    with open(fn, "rb") as f:
        events = json.load(f)
    data[name] = {"fn": os.path.basename(fn), "events": events}

from pprint import pprint
pprint(data)

here = os.path.dirname(__file__)
web_root = os.path.join(here, "web")
lib_root = os.path.join(here, "node_modules")
if not os.path.isdir(lib_root):
    print("Cannot find 'd3' and 'd3-tip' in misc/node_modules/")
    print("Please run 'npm install d3 d3-tip zepto' from the misc/ directory.")
    sys.exit(1)

def web():
    # set up a server that serves web/ at the root, plus a /data.json built
    # from {timeline}. Quit when it fetches /done .
    from twisted.web import resource, static, server
    from twisted.internet import reactor, endpoints
    ep = endpoints.serverFromString(reactor, "tcp:8066:interface=127.0.0.1")
    root = static.File(web_root)
    root.putChild("data.json", static.Data(json.dumps(data).encode("utf-8"),
                                           "application/json"))
    root.putChild("lib", static.File(lib_root))
    class Shutdown(resource.Resource):
        def render_GET(self, request):
            #print("timeline ready, server shutting down")
            #reactor.stop()
            return "shutting down"
    root.putChild("done", Shutdown())
    site = server.Site(root)
    ep.listen(site)
    import webbrowser
    def launch_browser():
        webbrowser.open("http://localhost:%d/timeline.html" % 8066)
        print("browser opened, waiting for shutdown")
    reactor.callLater(0, launch_browser)
    reactor.run()
web()



from __future__ import print_function, unicode_literals

from twisted.trial import unittest

from unittest import mock

from .._wordlist import PGPWordList


class Completions(unittest.TestCase):
    def test_completions(self):
        wl = PGPWordList()
        gc = wl.get_completions
        self.assertEqual(gc("ar", 2), {"armistice-", "article-"})
        self.assertEqual(gc("armis", 2), {"armistice-"})
        self.assertEqual(gc("armistice", 2), {"armistice-"})
        lots = gc("armistice-", 2)
        self.assertEqual(len(lots), 256, lots)
        first = list(lots)[0]
        self.assert_(first.startswith("armistice-"), first)
        self.assertEqual(
            gc("armistice-ba", 2), {
                "armistice-baboon", "armistice-backfield",
                "armistice-backward", "armistice-banjo"
            })
        self.assertEqual(
            gc("armistice-ba", 3), {
                "armistice-baboon-", "armistice-backfield-",
                "armistice-backward-", "armistice-banjo-"
            })
        self.assertEqual(gc("armistice-baboon", 2), {"armistice-baboon"})
        self.assertEqual(gc("armistice-baboon", 3), {"armistice-baboon-"})
        self.assertEqual(gc("armistice-baboon", 4), {"armistice-baboon-"})


class Choose(unittest.TestCase):
    def test_choose_words(self):
        wl = PGPWordList()
        with mock.patch("os.urandom", side_effect=[b"\x04", b"\x10"]):
            self.assertEqual(wl.choose_words(2), "alkali-assume")



# no unicode_literals until twisted update
from click.testing import CliRunner
from twisted.application import internet, service
from twisted.internet import defer, endpoints, reactor, task
from twisted.python import log

from unittest import mock
from wormhole_mailbox_server.database import create_channel_db, create_usage_db
from wormhole_mailbox_server.server import make_server
from wormhole_mailbox_server.web import make_web_server
from wormhole_transit_relay.transit_server import Transit

from ..cli import cli
from ..transit import allocate_tcp_port


class MyInternetService(service.Service, object):
    # like StreamServerEndpointService, but you can retrieve the port
    def __init__(self, endpoint, factory):
        self.endpoint = endpoint
        self.factory = factory
        self._port_d = defer.Deferred()
        self._lp = None

    def startService(self):
        super(MyInternetService, self).startService()
        d = self.endpoint.listen(self.factory)

        def good(lp):
            self._lp = lp
            self._port_d.callback(lp.getHost().port)

        def bad(f):
            log.err(f)
            self._port_d.errback(f)

        d.addCallbacks(good, bad)

    @defer.inlineCallbacks
    def stopService(self):
        if self._lp:
            yield self._lp.stopListening()

    def getPort(self):  # only call once!
        return self._port_d


class ServerBase:
    @defer.inlineCallbacks
    def setUp(self):
        yield self._setup_relay(None)

    @defer.inlineCallbacks
    def _setup_relay(self, error, advertise_version=None):
        self.sp = service.MultiService()
        self.sp.startService()
        # need to talk to twisted team about only using unicode in
        # endpoints.serverFromString
        db = create_channel_db(":memory:")
        self._usage_db = create_usage_db(":memory:")
        self._rendezvous = make_server(
            db,
            advertise_version=advertise_version,
            signal_error=error,
            usage_db=self._usage_db)
        ep = endpoints.TCP4ServerEndpoint(reactor, 0, interface="127.0.0.1")
        site = make_web_server(self._rendezvous, log_requests=False)
        # self._lp = yield ep.listen(site)
        s = MyInternetService(ep, site)
        s.setServiceParent(self.sp)
        self.rdv_ws_port = yield s.getPort()
        self._relay_server = s
        # self._rendezvous = s._rendezvous
        self.relayurl = u"ws://127.0.0.1:%d/v1" % self.rdv_ws_port
        # ws://127.0.0.1:%d/wormhole-relay/ws

        self.transitport = allocate_tcp_port()
        ep = endpoints.serverFromString(
            reactor, "tcp:%d:interface=127.0.0.1" % self.transitport)
        self._transit_server = f = Transit(
            blur_usage=None, log_file=None, usage_db=None)
        internet.StreamServerEndpointService(ep, f).setServiceParent(self.sp)
        self.transit = u"tcp:127.0.0.1:%d" % self.transitport

    @defer.inlineCallbacks
    def tearDown(self):
        # Unit tests that spawn a (blocking) client in a thread might still
        # have threads running at this point, if one is stuck waiting for a
        # message from a companion which has exited with an error. Our
        # relay's .stopService() drops all connections, which ought to
        # encourage those threads to terminate soon. If they don't, print a
        # warning to ease debugging.

        # XXX FIXME there's something in _noclobber test that's not
        # waiting for a close, I think -- was pretty relieably getting
        # unclean-reactor, but adding a slight pause here stops it...

        tp = reactor.getThreadPool()
        if not tp.working:
            yield self.sp.stopService()
            yield task.deferLater(reactor, 0.1, lambda: None)
            defer.returnValue(None)
        # disconnect all callers
        d = defer.maybeDeferred(self.sp.stopService)
        # wait a second, then check to see if it worked
        yield task.deferLater(reactor, 1.0, lambda: None)
        if len(tp.working):
            log.msg("wormhole.test.common.ServerBase.tearDown:"
                    " I was unable to convince all threads to exit.")
            tp.dumpStats()
            print("tearDown warning: threads are still active")
            print("This test will probably hang until one of the"
                  " clients gives up of their own accord.")
        else:
            log.msg("wormhole.test.common.ServerBase.tearDown:"
                    " I convinced all threads to exit.")
        yield d


def config(*argv):
    r = CliRunner()
    with mock.patch("wormhole.cli.cli.go") as go:
        res = r.invoke(cli.wormhole, argv, catch_exceptions=False)
        if res.exit_code != 0:
            print(res.exit_code)
            print(res.output)
            print(res)
            assert 0
        cfg = go.call_args[0][1]
    return cfg


@defer.inlineCallbacks
def poll_until(predicate):
    # return a Deferred that won't fire until the predicate is True
    while not predicate():
        d = defer.Deferred()
        reactor.callLater(0.001, d.callback, None)
        yield d



from ._rlcompleter import input_with_completion
from .wormhole import create

from . import _version
__version__ = _version.get_versions()['version']

__all__ = ["create", "input_with_completion", "__version__"]



from __future__ import print_function, unicode_literals

import io

from twisted.internet import defer
from twisted.internet.error import ConnectError
from twisted.trial import unittest

from unittest import mock

from .._interfaces import ITorManager
from ..errors import NoTorError
from ..tor_manager import SocksOnlyTor, get_tor


class X():
    pass


class Tor(unittest.TestCase):
    def test_no_txtorcon(self):
        with mock.patch("wormhole.tor_manager.txtorcon", None):
            self.failureResultOf(get_tor(None), NoTorError)

    def test_bad_args(self):
        f = self.failureResultOf(
            get_tor(None, launch_tor="not boolean"), TypeError)
        self.assertEqual(str(f.value), "launch_tor= must be boolean")

        f = self.failureResultOf(
            get_tor(None, tor_control_port=1234), TypeError)
        self.assertEqual(str(f.value), "tor_control_port= must be str or None")
        f = self.failureResultOf(
            get_tor(
                None, launch_tor=True, tor_control_port="tcp:127.0.0.1:1234"),
            ValueError)
        self.assertEqual(
            str(f.value),
            "cannot combine --launch-tor and --tor-control-port=")

    def test_launch(self):
        reactor = object()
        my_tor = X()  # object() didn't like providedBy()
        launch_d = defer.Deferred()
        stderr = io.StringIO()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.launch",
                side_effect=launch_d) as launch:
            d = get_tor(reactor, launch_tor=True, stderr=stderr)
            self.assertNoResult(d)
            self.assertEqual(launch.mock_calls, [mock.call(reactor)])
            launch_d.callback(my_tor)
            tor = self.successResultOf(d)
            self.assertIs(tor, my_tor)
            self.assert_(ITorManager.providedBy(tor))
            self.assertEqual(
                stderr.getvalue(),
                " launching a new Tor process, this may take a while..\n")

    def test_connect(self):
        reactor = object()
        my_tor = X()  # object() didn't like providedBy()
        connect_d = defer.Deferred()
        stderr = io.StringIO()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.connect",
                side_effect=connect_d) as connect:
            with mock.patch(
                    "wormhole.tor_manager.clientFromString",
                    side_effect=["foo"]) as sfs:
                d = get_tor(reactor, stderr=stderr)
        self.assertEqual(sfs.mock_calls, [])
        self.assertNoResult(d)
        self.assertEqual(connect.mock_calls, [mock.call(reactor)])
        connect_d.callback(my_tor)
        tor = self.successResultOf(d)
        self.assertIs(tor, my_tor)
        self.assert_(ITorManager.providedBy(tor))
        self.assertEqual(stderr.getvalue(),
                         " using Tor via default control port\n")

    def test_connect_fails(self):
        reactor = object()
        connect_d = defer.Deferred()
        stderr = io.StringIO()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.connect",
                side_effect=connect_d) as connect:
            with mock.patch(
                    "wormhole.tor_manager.clientFromString",
                    side_effect=["foo"]) as sfs:
                d = get_tor(reactor, stderr=stderr)
        self.assertEqual(sfs.mock_calls, [])
        self.assertNoResult(d)
        self.assertEqual(connect.mock_calls, [mock.call(reactor)])

        connect_d.errback(ConnectError())
        tor = self.successResultOf(d)
        self.assertIsInstance(tor, SocksOnlyTor)
        self.assert_(ITorManager.providedBy(tor))
        self.assertEqual(tor._reactor, reactor)
        self.assertEqual(
            stderr.getvalue(),
            " unable to find default Tor control port, using SOCKS\n")

    def test_connect_custom_control_port(self):
        reactor = object()
        my_tor = X()  # object() didn't like providedBy()
        tcp = "PORT"
        ep = object()
        connect_d = defer.Deferred()
        stderr = io.StringIO()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.connect",
                side_effect=connect_d) as connect:
            with mock.patch(
                    "wormhole.tor_manager.clientFromString",
                    side_effect=[ep]) as sfs:
                d = get_tor(reactor, tor_control_port=tcp, stderr=stderr)
        self.assertEqual(sfs.mock_calls, [mock.call(reactor, tcp)])
        self.assertNoResult(d)
        self.assertEqual(connect.mock_calls, [mock.call(reactor, ep)])
        connect_d.callback(my_tor)
        tor = self.successResultOf(d)
        self.assertIs(tor, my_tor)
        self.assert_(ITorManager.providedBy(tor))
        self.assertEqual(stderr.getvalue(),
                         " using Tor via control port at PORT\n")

    def test_connect_custom_control_port_fails(self):
        reactor = object()
        tcp = "port"
        ep = object()
        connect_d = defer.Deferred()
        stderr = io.StringIO()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.connect",
                side_effect=connect_d) as connect:
            with mock.patch(
                    "wormhole.tor_manager.clientFromString",
                    side_effect=[ep]) as sfs:
                d = get_tor(reactor, tor_control_port=tcp, stderr=stderr)
        self.assertEqual(sfs.mock_calls, [mock.call(reactor, tcp)])
        self.assertNoResult(d)
        self.assertEqual(connect.mock_calls, [mock.call(reactor, ep)])

        connect_d.errback(ConnectError())
        self.failureResultOf(d, ConnectError)
        self.assertEqual(stderr.getvalue(), "")


class SocksOnly(unittest.TestCase):
    def test_tor(self):
        reactor = object()
        sot = SocksOnlyTor(reactor)
        fake_ep = object()
        with mock.patch(
                "wormhole.tor_manager.txtorcon.TorClientEndpoint",
                return_value=fake_ep) as tce:
            ep = sot.stream_via("host", "port")
        self.assertIs(ep, fake_ep)
        self.assertEqual(tce.mock_calls, [
            mock.call(
                "host",
                "port",
                socks_endpoint=None,
                tls=False,
                reactor=reactor)
        ])



from __future__ import print_function, unicode_literals

from unittest import mock

from twisted.trial import unittest

from .._key import derive_key, derive_phase_key, encrypt_data, decrypt_data
from ..util import bytes_to_hexstr, hexstr_to_bytes


class Derive(unittest.TestCase):
    def test_derive_errors(self):
        self.assertRaises(TypeError, derive_key, 123, b"purpose")
        self.assertRaises(TypeError, derive_key, b"key", 123)
        self.assertRaises(TypeError, derive_key, b"key", b"purpose", "not len")

    def test_derive_key(self):
        m = "588ba9eef353778b074413a0140205d90d7479e36e0dd4ee35bb729d26131ef1"
        main = hexstr_to_bytes(m)

        dk1 = derive_key(main, b"purpose1")
        self.assertEqual(bytes_to_hexstr(dk1),
                         "835b5df80ce9ca46908e8524fb308649"
                         "122cfbcefbeaa7e65061c6ef08ee1b2a")

        dk2 = derive_key(main, b"purpose2", 10)
        self.assertEqual(bytes_to_hexstr(dk2), "f2238e84315b47eb6279")

    def test_derive_phase_key(self):
        m = "588ba9eef353778b074413a0140205d90d7479e36e0dd4ee35bb729d26131ef1"
        main = hexstr_to_bytes(m)

        dk11 = derive_phase_key(main, "side1", "phase1")
        self.assertEqual(bytes_to_hexstr(dk11),
                         "3af6a61d1a111225cc8968c6ca6265ef"
                         "e892065c3ab46de79dda21306b062990")

        dk12 = derive_phase_key(main, "side1", "phase2")
        self.assertEqual(bytes_to_hexstr(dk12),
                         "88a1dd12182d989ff498022a9656d1e2"
                         "806f17328d8bf5d8d0c9753e4381a752")

        dk21 = derive_phase_key(main, "side2", "phase1")
        self.assertEqual(bytes_to_hexstr(dk21),
                         "a306627b436ec23bdae3af8fa90c9ac9"
                         "27780d86be1831003e7f617c518ea689")

        dk22 = derive_phase_key(main, "side2", "phase2")
        self.assertEqual(bytes_to_hexstr(dk22),
                         "bf99e3e16420f2dad33f9b1ccb0be146"
                         "2b253d639dacdb50ed9496fa528d8758")


class Encrypt(unittest.TestCase):
    def test_encrypt(self):
        k = "ddc543ef8e4629a603d39dd0307a51bb1e7adb9cb259f6b085c91d0842a18679"
        key = hexstr_to_bytes(k)
        plaintext = hexstr_to_bytes("edc089a518219ec1cee184e89d2d37af")
        self.assertEqual(len(plaintext), 16)
        nonce = hexstr_to_bytes("2d5e43eb465aa42e750f991e425bee48"
                                "5f06abad7e04af80")
        self.assertEqual(len(nonce), 24)
        with mock.patch("nacl.utils.random", return_value=nonce):
            encrypted = encrypt_data(key, plaintext)
        self.assertEqual(len(encrypted), 24 + 16 + 16)
        self.assertEqual(bytes_to_hexstr(encrypted),
                         "2d5e43eb465aa42e750f991e425bee48"
                         "5f06abad7e04af80fe318e39d0e4ce93"
                         "2d2b54b300c56d2cda55ee5f0488d63e"
                         "b1d5f76f7919a49a")

    def test_decrypt(self):
        k = "ddc543ef8e4629a603d39dd0307a51bb1e7adb9cb259f6b085c91d0842a18679"
        key = hexstr_to_bytes(k)
        encrypted = hexstr_to_bytes("2d5e43eb465aa42e750f991e425bee48"
                                    "5f06abad7e04af80fe318e39d0e4ce93"
                                    "2d2b54b300c56d2cda55ee5f0488d63e"
                                    "b1d5f76f7919a49a")

        decrypted = decrypt_data(key, encrypted)
        self.assertEqual(len(decrypted), len(encrypted) - 24 - 16)
        self.assertEqual(bytes_to_hexstr(decrypted),
                         "edc089a518219ec1cee184e89d2d37af")



class _Role(object):
    def __init__(self, which):
        self._which = which

    def __repr__(self):
        return "Role(%s)" % self._which


LEADER, FOLLOWER = _Role("LEADER"), _Role("FOLLOWER")



from __future__ import print_function, unicode_literals
from unittest import mock
from zope.interface import alsoProvides
from twisted.trial import unittest
from ..._interfaces import IDilationManager
from ..._dilation.connection import Open, Data, Close
from ..._dilation.inbound import (Inbound, DuplicateOpenError,
                                  DataForMissingSubchannelError,
                                  CloseForMissingSubchannelError)


def make_inbound():
    m = mock.Mock()
    alsoProvides(m, IDilationManager)
    host_addr = object()
    i = Inbound(m, host_addr)
    return i, m, host_addr


class InboundTest(unittest.TestCase):
    def test_seqnum(self):
        i, m, host_addr = make_inbound()
        r1 = Open(scid=513, seqnum=1)
        r2 = Data(scid=513, seqnum=2, data=b"")
        r3 = Close(scid=513, seqnum=3)
        self.assertFalse(i.is_record_old(r1))
        self.assertFalse(i.is_record_old(r2))
        self.assertFalse(i.is_record_old(r3))

        i.update_ack_watermark(r1.seqnum)
        self.assertTrue(i.is_record_old(r1))
        self.assertFalse(i.is_record_old(r2))
        self.assertFalse(i.is_record_old(r3))

        i.update_ack_watermark(r2.seqnum)
        self.assertTrue(i.is_record_old(r1))
        self.assertTrue(i.is_record_old(r2))
        self.assertFalse(i.is_record_old(r3))

    def test_open_data_close(self):
        i, m, host_addr = make_inbound()
        scid1 = b"scid"
        scid2 = b"scXX"
        c = mock.Mock()
        lep = mock.Mock()
        i.set_listener_endpoint(lep)
        i.use_connection(c)
        sc1 = mock.Mock()
        peer_addr = object()
        with mock.patch("wormhole._dilation.inbound.SubChannel",
                        side_effect=[sc1]) as sc:
            with mock.patch("wormhole._dilation.inbound._SubchannelAddress",
                            side_effect=[peer_addr]) as sca:
                i.handle_open(scid1)
        self.assertEqual(lep.mock_calls, [mock.call._got_open(sc1, peer_addr)])
        self.assertEqual(sc.mock_calls, [mock.call(scid1, m, host_addr, peer_addr)])
        self.assertEqual(sca.mock_calls, [mock.call(scid1)])
        lep.mock_calls[:] = []

        # a subsequent duplicate OPEN should be ignored
        with mock.patch("wormhole._dilation.inbound.SubChannel",
                        side_effect=[sc1]) as sc:
            with mock.patch("wormhole._dilation.inbound._SubchannelAddress",
                            side_effect=[peer_addr]) as sca:
                i.handle_open(scid1)
        self.assertEqual(lep.mock_calls, [])
        self.assertEqual(sc.mock_calls, [])
        self.assertEqual(sca.mock_calls, [])
        self.flushLoggedErrors(DuplicateOpenError)

        i.handle_data(scid1, b"data")
        self.assertEqual(sc1.mock_calls, [mock.call.remote_data(b"data")])
        sc1.mock_calls[:] = []

        i.handle_data(scid2, b"for non-existent subchannel")
        self.assertEqual(sc1.mock_calls, [])
        self.flushLoggedErrors(DataForMissingSubchannelError)

        i.handle_close(scid1)
        self.assertEqual(sc1.mock_calls, [mock.call.remote_close()])
        sc1.mock_calls[:] = []

        i.handle_close(scid2)
        self.assertEqual(sc1.mock_calls, [])
        self.flushLoggedErrors(CloseForMissingSubchannelError)

        # after the subchannel is closed, the Manager will notify Inbound
        i.subchannel_closed(scid1, sc1)

        i.stop_using_connection()

    def test_control_channel(self):
        i, m, host_addr = make_inbound()
        lep = mock.Mock()
        i.set_listener_endpoint(lep)

        scid0 = b"scid"
        sc0 = mock.Mock()
        i.set_subchannel_zero(scid0, sc0)

        # OPEN on the control channel identifier should be ignored as a
        # duplicate, since the control channel is already registered
        sc1 = mock.Mock()
        peer_addr = object()
        with mock.patch("wormhole._dilation.inbound.SubChannel",
                        side_effect=[sc1]) as sc:
            with mock.patch("wormhole._dilation.inbound._SubchannelAddress",
                            side_effect=[peer_addr]) as sca:
                i.handle_open(scid0)
        self.assertEqual(lep.mock_calls, [])
        self.assertEqual(sc.mock_calls, [])
        self.assertEqual(sca.mock_calls, [])
        self.flushLoggedErrors(DuplicateOpenError)

        # and DATA to it should be delivered correctly
        i.handle_data(scid0, b"data")
        self.assertEqual(sc0.mock_calls, [mock.call.remote_data(b"data")])
        sc0.mock_calls[:] = []

    def test_pause(self):
        i, m, host_addr = make_inbound()
        c = mock.Mock()
        lep = mock.Mock()
        i.set_listener_endpoint(lep)

        # add two subchannels, pause one, then add a connection
        scid1 = b"sci1"
        scid2 = b"sci2"
        sc1 = mock.Mock()
        sc2 = mock.Mock()
        peer_addr = object()
        with mock.patch("wormhole._dilation.inbound.SubChannel",
                        side_effect=[sc1, sc2]):
            with mock.patch("wormhole._dilation.inbound._SubchannelAddress",
                            return_value=peer_addr):
                i.handle_open(scid1)
                i.handle_open(scid2)
        self.assertEqual(c.mock_calls, [])

        i.subchannel_pauseProducing(sc1)
        self.assertEqual(c.mock_calls, [])
        i.subchannel_resumeProducing(sc1)
        self.assertEqual(c.mock_calls, [])
        i.subchannel_pauseProducing(sc1)
        self.assertEqual(c.mock_calls, [])

        i.use_connection(c)
        self.assertEqual(c.mock_calls, [mock.call.pauseProducing()])
        c.mock_calls[:] = []

        i.subchannel_resumeProducing(sc1)
        self.assertEqual(c.mock_calls, [mock.call.resumeProducing()])
        c.mock_calls[:] = []

        # consumers aren't really supposed to do this, but tolerate it
        i.subchannel_resumeProducing(sc1)
        self.assertEqual(c.mock_calls, [])

        i.subchannel_pauseProducing(sc1)
        self.assertEqual(c.mock_calls, [mock.call.pauseProducing()])
        c.mock_calls[:] = []
        i.subchannel_pauseProducing(sc2)
        self.assertEqual(c.mock_calls, [])  # was already paused

        # tolerate duplicate pauseProducing
        i.subchannel_pauseProducing(sc2)
        self.assertEqual(c.mock_calls, [])

        # stopProducing is treated like a terminal resumeProducing
        i.subchannel_stopProducing(sc1)
        self.assertEqual(c.mock_calls, [])
        i.subchannel_stopProducing(sc2)
        self.assertEqual(c.mock_calls, [mock.call.resumeProducing()])
        c.mock_calls[:] = []



import unicodedata

from twisted.trial import unittest

from unittest import mock

from .. import util


class Utils(unittest.TestCase):
    def test_to_bytes(self):
        b = util.to_bytes("abc")
        self.assertIsInstance(b, type(b""))
        self.assertEqual(b, b"abc")

        A = unicodedata.lookup("LATIN SMALL LETTER A WITH DIAERESIS")
        b = util.to_bytes(A + "bc")
        self.assertIsInstance(b, type(b""))
        self.assertEqual(b, b"\xc3\xa4\x62\x63")

    def test_bytes_to_hexstr(self):
        b = b"\x00\x45\x91\xfe\xff"
        hexstr = util.bytes_to_hexstr(b)
        self.assertIsInstance(hexstr, type(""))
        self.assertEqual(hexstr, "004591feff")

    def test_hexstr_to_bytes(self):
        hexstr = "004591feff"
        b = util.hexstr_to_bytes(hexstr)
        hexstr = util.bytes_to_hexstr(b)
        self.assertIsInstance(b, type(b""))
        self.assertEqual(b, b"\x00\x45\x91\xfe\xff")

    def test_dict_to_bytes(self):
        d = {"a": "b"}
        b = util.dict_to_bytes(d)
        self.assertIsInstance(b, type(b""))
        self.assertEqual(b, b'{"a": "b"}')

    def test_bytes_to_dict(self):
        b = b'{"a": "b", "c": 2}'
        d = util.bytes_to_dict(b)
        self.assertIsInstance(d, dict)
        self.assertEqual(d, {"a": "b", "c": 2})


class Space(unittest.TestCase):
    def test_free_space(self):
        free = util.estimate_free_space(".")
        self.assert_(
            isinstance(free, (int, type(None))), repr(free))
        # some platforms (I think the VMs used by travis are in this
        # category) return 0, and windows will return None, so don't assert
        # anything more specific about the return value

    def test_no_statvfs(self):
        # this mock.patch fails on windows, which is sad because windows is
        # the one platform that the code under test was supposed to help with
        try:
            with mock.patch("os.statvfs", side_effect=AttributeError()):
                self.assertEqual(util.estimate_free_space("."), None)
        except AttributeError:  # raised by mock.get_original()
            pass
